/*
**	$Id:$
**
**	Copyright (C) 1999 Bernardo Innocenti
**	All rights reserved.
**
**	Use 4 chars wide TABs to read this file
**
**	GetGadgetBox() computes the actual IBox where a gadget exists in a window.
**	The special cases it handles are all the REL#? (relative positioning flags).
**
**	This function returns the gadget size in the provided
**	IBox structure, computing the values from the coordinates
**	in the gadget structure and the container (window or requester)
**	where it lives.
*/

#include <intuition/intuition.h>

#include <CompilerSpecific.h>
#include <DebugMacros.h>
#include <BoopsiLib.h>


void GetGadgetBox(struct GadgetInfo *ginfo, struct ExtGadget *g, struct IBox *box)
{
	ASSERT_VALID_PTR(g)
	ASSERT_VALID_PTR(ginfo)
	ASSERT_VALID_PTR(box)

	DB2(if (g->Flags & GFLG_EXTENDED)
		DBPRINTF("GetGadgetBox(): GFLG_EXTENDED is set\n");)
	DB2(if ((g->Flags & GFLG_EXTENDED) && (g->MoreFlags & GMORE_BOUNDS))
		DBPRINTF("GetGadgetBox(): Gadget has valid bounds\n");)


	box->Left = g->LeftEdge;
	if (g->Flags & GFLG_RELRIGHT)
		box->Left += ginfo->gi_Domain.Width - 1;

	box->Top = g->TopEdge;
	if (g->Flags & GFLG_RELBOTTOM)
		box->Top += ginfo->gi_Domain.Height - 1;

	box->Width = g->Width;
	if (g->Flags & GFLG_RELWIDTH)
		box->Width += ginfo->gi_Domain.Width;

	box->Height = g->Height;
	if (g->Flags & GFLG_RELHEIGHT)
		box->Height += ginfo->gi_Domain.Height;

	DB2(DBPRINTF("GetGadgetBox(): Left = %ld, Top = %ld, Width = %ld, Height = %ld\n",
		box->Left, box->Top, box->Width, box->Height);)
}



/* Useful macro missing in <intuition/imageclass.h> */
#define GADGET_BOUNDS(g) ((struct IBox *)&((struct ExtGadget *)(g))->BoundsLeftEdge)


void GetGadgetBounds(struct GadgetInfo *ginfo, struct ExtGadget *g, struct IBox *bounds)
{
	struct IBox *box;

	ASSERT_VALID_PTR(g)
	ASSERT_VALID_PTR(ginfo)
	ASSERT_VALID_PTR(bounds)


	box = ((g->Flags & GFLG_EXTENDED) && (g->MoreFlags & GMORE_BOUNDS)) ?
		GADGET_BOUNDS(g) : GADGET_BOX(g);

	memcpy(bounds, box, sizeof(struct IBox));

	if (g->Flags & GFLG_RELRIGHT)
		bounds->Left += ginfo->gi_Domain.Width - 1;

	if (g->Flags & GFLG_RELBOTTOM)
		bounds->Top += ginfo->gi_Domain.Height - 1;

	if (g->Flags & GFLG_RELWIDTH)
		bounds->Width += ginfo->gi_Domain.Width;

	if (g->Flags & GFLG_RELHEIGHT)
		bounds->Height += ginfo->gi_Domain.Height;
}
