/*
**	$Id: ListBoxClass.c,v 1.3 1999/01/30 13:25:03 bernie Exp $
**
**	Copyright (C) 1997,98,99 Bernardo Innocenti (<bernardo.innocenti@usa.net>)
**	All rights reserved.
**
**	Use 4 chars wide TABs to read this file
**
**	GadTools-like `boopsi' ListView group class
*/

#define USE_BUILTIN_MATH
#define INTUI_V36_NAMES_ONLY
#define INTUITION_IOBSOLETE_H
#define __USE_SYSBASE
#define  CLIB_ALIB_PROTOS_H		/* Avoid dupe defines of boopsi funcs */

#include <exec/types.h>
#include <exec/memory.h>
#include <intuition/intuition.h>
#include <intuition/intuitionbase.h>
#include <intuition/classes.h>
#include <intuition/gadgetclass.h>
#include <intuition/icclass.h>
#include <intuition/imageclass.h>

#include <proto/exec.h>
#include <proto/intuition.h>
#include <proto/utility.h>

#ifdef __STORM__
	#pragma header
#endif

#define LV_GADTOOLS_STUFF
#include <gadgets/ListBoxClass.h>

#include <CompilerSpecific.h>
#include <DebugMacros.h>
#include <BoopsiStubs.h>
#include <BoopsiLib.h>



/* Sliders dimensions
 */
#define VSLIDER_WIDTH	18
#define HSLIDER_HEIGHT	18


/* Per-object instance data */
struct LBData
{
	/* struct Gadget *ThisGadget; (not used) */

	/* Group children */
	Object *ListView;
	Object *HSlider;
	Object *VSlider;
	Object *LVToVSliderIC;
	Object *LVToHSliderIC;

	Object *Model;			/* The ic object that makes our children talk to each other */
	Object *Frame;			/* The frame to put around the child objects */

	/* Frame size */
	LONG	FrameWidth, FrameHeight;

	/* These two have the same meaning, but we keep both updated
	 * because the Rectangle structure (MinX, MinY, MaxX, MaxY)
	 * is more handy in some cases, while the IBox structure
	 * (Left/Top/Width/Height) is best for other cases.
	 */
	struct IBox		 GBox;
	struct Rectangle GRect;
};


/* Global class data */
struct LBClassData
{
	Class				*SliderBarGClass;
	struct ClassLibrary	*SliderBarBase;
};



#define SLIDERBARGCLASS_PTR ( ((struct LBClassData *)(cl->cl_UserData))->SliderBarGClass )



extern Class *ListViewClass;


/* Local function prototypes */
static void		LB_GMLayout		(Class *cl, struct Gadget *g, struct gpLayout *msg);
static ULONG	LB_OMSet		(Class *cl, struct Gadget *g, struct opUpdate *msg);
static ULONG	LB_OMGet		(Class *cl, struct Gadget *g, struct opGet *msg);
static ULONG	LB_OMNew		(Class *cl, struct Gadget *g, struct opSet *msg);
static void		LB_OMDispose	(Class *cl, struct Gadget *g, Msg msg);

static void		CreateVSlider	(Class *cl, struct Gadget *g, struct LBData *lb, struct DrawInfo *dri);
static void		CreateHSlider	(Class *cl, struct Gadget *g, struct LBData *lb, struct DrawInfo *dri);
static void		DeleteVSlider	(struct Gadget *g, struct LBData *lb);
static void		DeleteHSlider	(struct Gadget *g, struct LBData *lb);




/* Attribute translations for object interconnections */

static LONG MapLVToHSlider[] =
{
	LVA_PixelLeft,		PGA_Top,
	LVA_PixelWidth,		PGA_Total,
	LVA_PixelHVisible,	PGA_Visible,
	TAG_DONE
};

static LONG MapHSliderToLV[] =
{
	PGA_Top,			LVA_PixelLeft,
	TAG_DONE
};

static LONG MapLVToVSlider[] =
{
	LVA_PixelTop,		PGA_Top,
	LVA_PixelHeight,	PGA_Total,
	LVA_PixelVVisible,	PGA_Visible,
	TAG_DONE
};

static LONG MapVSliderToLV[] =
{
	PGA_Top,	LVA_PixelTop,
	TAG_DONE
};



static ULONG HOOKCALL LBDispatcher(
	REG(a0, Class *cl),
	REG(a2, struct Gadget *g),
	REG(a1, Msg msg))
{
	ASSERT_VALID_PTR(cl)
	ASSERT_VALID_PTR(g)
	ASSERT_VALID_PTR(msg)

	switch (msg->MethodID)
	{
		case GM_LAYOUT:
			/* This method is only supported on V39 and above */
			LB_GMLayout(cl, g, (struct gpLayout *)msg);
			return TRUE;

		case OM_SET:
		case OM_UPDATE:
			return LB_OMSet(cl, g, (struct opUpdate *)msg);

		case OM_GET:
			return LB_OMGet(cl, g, (struct opGet *)msg);

		case OM_NEW:
			return LB_OMNew(cl, g, (struct opSet *)msg);

		case OM_DISPOSE:
			LB_OMDispose(cl, g, msg);
			return TRUE;

		default:
			/* Unsupported method: let our superclass's dispatcher take
			 * a look at it. This includes all gadget methods sent
			 * by Intuition: GM_RENDER, GM_HANDLEINPUT, GM_GOACTIVE and
			 * GM_GOINACTIVE. These methods are automatically forwarded
			 * to our child gadgets by the groupgclass.
			 */
			DB2 (DBPRINTF("ListBoxClass: passing unknown method 0x%lx to superclass\n", msg->MethodID);)
			return DoSuperMethodA (cl, (Object *)g, msg);
	}
}



static void LB_GMLayout(Class *cl, struct Gadget *g, struct gpLayout *msg)
{
	struct LBData *lb = (struct LBData *) INST_DATA(cl, (Object *)g);

	DB(DBPRINTF("ListBoxClass: GM_LAYOUT\n");)
	ASSERT_VALID_PTR(lb)


	/* Collect new gadget size */
	GetGadgetBox(msg->gpl_GInfo, (struct ExtGadget *)g, &lv->GBox);
	IBoxToRect(&lv->GBox, &lv->GRect);


	/* Size our children accordingly */
	SetAttrs(lb->ListView,
		GA_Left,	lb->GBox.Left + lb->FrameWidth / 2,
		GA_Top,		lb->GBox.Top + lb->FrameHeight / 2,
		GA_Width,	lb->GBox.Width - lb->FrameWidth - VSLIDER_WIDTH,
		GA_Height,	lb->GBox.Height - lb->FrameHeight - HSLIDER_HEIGHT,
		TAG_DONE);

	if (lb->VSlider)
		SetAttrs(lb->VSlider,
			GA_Left,	lb->GRect.MaxX - VSLIDER_WIDTH + 1,
			GA_Top,		lb->GBox.Top,
			GA_Width,	VSLIDER_WIDTH,
			GA_Height,	lb->GBox.Height,
			TAG_DONE);


	/* NOTE: it seems that the groupgclass does not forward GM_LAYOUT
	 * to its children, so we must handle this here.
	 */
	DoMethodA(lb->ListView, (Msg)msg);
	if (lb->VSlider)	DoMethodA(lb->VSlider, (Msg)msg);
}



static ULONG LB_OMSet(Class *cl, struct Gadget *g, struct opUpdate *msg)
{
	struct LBData *lb = (struct LBData *) INST_DATA(cl, (Object *)g);

	DB2(DBPRINTF("ListBoxClass: OM_SET\n");)
	ASSERT_VALID_PTR(lb)

	if (lb->ListView)
		/* Forward attributes to our listview */
		DoMethodA (lb->ListView, (Msg)msg);

	/* Also forward to our superclass */
	return DoSuperMethodA (cl, (Object *)g, (Msg) msg);
}



static ULONG LB_OMGet(Class *cl, struct Gadget *g, struct opGet *msg)
{
	struct LBData *lb = (struct LBData *) INST_DATA(cl, (Object *)g);

	DB2(DBPRINTF("ListBoxClass: OM_GET\n");)
	ASSERT_VALID_PTR(lb)

	/* Forward this method to our listview */
	return DoMethodA(lb->ListView, (Msg)msg);
}



static ULONG LB_OMNew(Class *cl, struct Gadget *g, struct opSet *msg)
{
	struct LBData	*lb;
	struct DrawInfo	*dri;


	DB2(DBPRINTF("ListBoxClass: OM_NEW\n");)

	if (g = (struct Gadget *)DoSuperMethodA(cl, (Object *)g, (Msg)msg))
	{
		/* Set the GMORE_SCROLLRASTER flag */
		if (g->Flags & GFLG_EXTENDED)
		{
			DB(DBPRINTF("ListBoxClass: OM_NEW: Setting GMORE_SCROLLRASTER\n");)
			((struct ExtGadget *)g)->MoreFlags |= GMORE_SCROLLRASTER;
		}

		lb = (struct LBData *) INST_DATA(cl, (Object *)g);
		ASSERT_VALID_PTR(lb)


		/* Clear the object instance */
		memset(lb, 0, sizeof (struct LBData));


		/* Store a pointer to this object (a Gadget) in the class private
		 * instance. This way we can avoid passing it along to all functions.
		 */
		/* lb->ThisGadget = g; (not used) */


		/* May be NULL */
		dri = (struct DrawInfo *) GetTagData(GA_DrawInfo, NULL, msg->ops_AttrList);


		/* Create a frame to put around us */
		if (lb->Frame = NewObject(NULL, FRAMEICLASS,
			IA_EdgesOnly,	TRUE,
			IA_FrameType,	FRAME_BUTTON,
			TAG_DONE))
		{
			struct IBox FrameBox, ContentsBox = { 0, 0, 0, 0 };

			/* Ask the frame about its nominal frame width and height */
			DoMethod((Object *)lb->Frame, IM_FRAMEBOX, &ContentsBox, &FrameBox, dri, 0);

			/* Remember it later */
			lb->FrameWidth = FrameBox.Width;
			lb->FrameHeight = FrameBox.Height;
		}

		/* Create a model object. This will be the core of the boopsi attributes
		 * network used by the sub-objects to talk each other. We pass our
		 * initalization tags to the model so it will pick up the correct
		 * ICA_TARGET and ICA_MAP, if specified with NewObject().
		 */
		if (lb->Model = NewObjectA(NULL, MODELCLASS, NULL))
		{
			/* Create the ListView and pass all creation time attributes to it.
			 * Note that any GA_#? attributes are also passed to the listview,
			 * so it will have the same size of its container.
			 */
			if (lb->ListView = NewObject(ListViewClass, NULL,
				GA_Image,		lb->Frame,
				TAG_MORE,		msg->ops_AttrList))
			{
				/* From now on, the groupgclass will dispose this object for us */
				DoMethod((Object *)g, OM_ADDMEMBER, lb->ListView);

				/* Connect Model to ListView */
				{
					APTR icobject;

					if (icobject = NewObject(NULL, ICCLASS,
						ICA_TARGET,	lb->ListView,
						TAG_DONE))
						if (!DoMethod(lb->Model, OM_ADDMEMBER, icobject))
							DisposeObject(icobject);
				}

				/* Connect ListView to Model */
				SetAttrs(lb->ListView,
					ICA_TARGET, lb->Model,
					TAG_DONE);

				/* Add sliders */
				CreateVSlider(cl, g, lb, dri);
				CreateHSlider(cl, g, lb, dri);

				/* Set the gadget width and height because the groupgclass
				 * always forces them to 0 on creation.
				 */
				{
					struct TagItem *tag;

					if (tag = FindTagItem(GA_RelWidth, msg->ops_AttrList))
						SetAttrs(g,
							GA_RelWidth, tag->ti_Data,
							TAG_DONE);
					else
						SetAttrs (g,
							GA_Width, GetTagData(GA_Width, g->Width, msg->ops_AttrList),
							TAG_DONE);

					if (tag = FindTagItem(GA_RelHeight, msg->ops_AttrList))
						SetAttrs(g,
							GA_RelHeight, tag->ti_Data,
							TAG_DONE);
					else
						SetAttrs(g,
							GA_Height, GetTagData (GA_Height, g->Height, msg->ops_AttrList),
							TAG_DONE);

					DB2(DBPRINTF("ListBoxClass: OM_NEW: size set to L=%ld T=%ld W=%ld H=%ld\n",
						g->LeftEdge, g->TopEdge, g->Width, g->Height);)
				}

				/* TODO: Handle creation-time attributes */

				return (ULONG)g;	/* Return newly created istance */
			}
		}


		/* Dispose object without disturbing the dispatchers of our sub-classes, if any */
		CoerceMethod(cl, (Object *)g, OM_DISPOSE);
	}

	return 0;	/* Fail */
}



static void LB_OMDispose(Class *cl, struct Gadget *g, Msg msg)
{
	struct LBData	*lb = (struct LBData *) INST_DATA(cl, (Object *)g);

	DB(DBPRINTF("ListBoxClass: OM_DISPOSE\n");)
	ASSERT_VALID_PTR(lb)


	DeleteHSlider(g, lb);
	DeleteVSlider(g, lb);

	/* Dispose the child objects that aren't freed automatically by the modelclass.
	 * Note that here we are disposing the Frame before we dispose the gadgets that
	 * may be still using it. Everything will go well unless those classes will
	 * access the frame within their OM_DISPOSE methods (very silly).
	 */
	DisposeObject(lb->Model);
	DisposeObject(lb->Frame);

	/* Our superclass will cleanup everything else now */
	DoSuperMethodA(cl, (Object *)g, (Msg) msg);

	/* From now on, our instance data is no longer available */
}



static void CreateVSlider(Class *cl, struct Gadget *g, struct LBData *lb, struct DrawInfo *dri)
{
	if (lb->VSlider = NewObject(SLIDERBARGCLASS_PTR, NULL,
		GA_ID,			g->GadgetID,	/* Same as our ID */
		GA_DrawInfo,	dri,
		GA_LabelImage,	lb->Frame,
		PGA_Freedom,	FREEVERT,
		PGA_NewLook,	TRUE,
		PGA_Borderless,	TRUE,
		LAYOUTA_Orientation, LORIENT_VERT,
		ICA_TARGET,		lb->Model,
		ICA_MAP,		MapVSliderToLV,
		TAG_DONE))
	{
		/* From now on, the groupgclass will dispose this object for us */
		DoMethod((Object *)g, OM_ADDMEMBER, lb->VSlider);

		/* Connect Model to ListView */
		SetAttrs(lb->Model,
			ICA_TARGET, lb->ListView,
			TAG_DONE);

		/* Connect VSlider to Model */
		if (lb->LVToVSliderIC = NewObject(NULL, ICCLASS,
			ICA_TARGET,	lb->VSlider,
			ICA_MAP,	MapLVToVSlider,
			TAG_DONE))
			if (!DoMethod(lb->Model, OM_ADDMEMBER, lb->LVToVSliderIC))
				DisposeObject(lb->LVToVSliderIC);
	}
}



static void CreateHSlider(Class *cl, struct Gadget *g, struct LBData *lb, struct DrawInfo *dri)
{
	/* TODO */
}


static void DeleteVSlider(struct Gadget *g, struct LBData *lb)
{
	if (lb->LVToVSliderIC)
	{
		ASSERT_VALID_PTR(lb->LVToVSliderIC)
		ASSERT_VALID_PTR(lb->Model)

		DoMethod (lb->Model, OM_REMMEMBER, lb->LVToVSliderIC);
		DisposeObject (lb->LVToVSliderIC);
		lb->LVToVSliderIC = NULL;
	}

	if (lb->VSlider)
	{
		ASSERT_VALID_PTR(lb->VSlider)

		DoMethod ((Object *)g, OM_REMMEMBER, lb->VSlider);
		DisposeObject (lb->VSlider);
		lb->VSlider = NULL;
	}
}



static void DeleteHSlider(struct Gadget *g, struct LBData *lb)
{
	/* TODO */
}



Class *MakeListBoxClass(void)
{
	Class *class;
	struct LBClassData *classdata;

	if (class = MakeClass(NULL, GROUPGCLASS, NULL, sizeof (struct LBData), 0))
	{
		class->cl_Dispatcher.h_Entry = (ULONG (*)()) LBDispatcher;

		/* Allocate storage for global class data */
		if (classdata = AllocMem(sizeof (struct LBClassData), MEMF_PUBLIC | MEMF_CLEAR))
		{
			class->cl_UserData = (ULONG) classdata;

			if (classdata->SliderBarBase = (struct ClassLibrary *)
				OpenLibrary("gadgets/sliderbar.gadget", 0))
			{
				classdata->SliderBarGClass = classdata->SliderBarBase->cl_Class;
				ASSERT_VALID_PTR(classdata->SliderBarGClass);

				return class;
			}
		}

		FreeListBoxClass(class);
	}

	return NULL;
}



BOOL FreeListBoxClass(Class *class)
{
	struct LBClassData *classdata;

	if (class)
	{
		ASSERT_VALID_PTR(class)

		classdata = (struct LBClassData *)class->cl_UserData;

		/* Try to remove the class */
		if (FreeClass(class))
		{
			if (classdata)
			{
				ASSERT_VALID_PTR(classdata)
				ASSERT_VALID_PTR_OR_NULL(classdata->SliderBarBase)

				/* Cleanup global class data */

				/* NULL is safe in CloseLibrary() since V36 */
				CloseLibrary((struct Library *)classdata->SliderBarBase);
				FreeMem(classdata, sizeof (struct LBClassData));
			}
			return TRUE;
		}
		return FALSE;
	}
	return TRUE;
}
