/*
**	$Id: ScrollButtonClass.c,v 1.1 1999/02/07 14:24:43 bernie Exp $
**
**	Copyright (C) 1998,99 Bernardo Innocenti
**
**	Use 4 chars wide TABs to read this file
**
**	The ScrollButtonClass is a simple subclass of the buttongclass
**	specialized for scroll buttons. See ScrollButtonClass.h for a detailed
**	description.
*/

#define USE_BUILTIN_MATH
#define INTUI_V36_NAMES_ONLY
#define __USE_SYSBASE
#define  CLIB_ALIB_PROTOS_H		/* Avoid dupe defs of boopsi funcs */


#include <intuition/gadgetclass.h>
#include <intuition/imageclass.h>
#include <proto/exec.h>
#include <proto/intuition.h>
#include <proto/utility.h>

#include <gadgets/ScrollButtonClass.h>

#include <CompilerSpecific.h>
#include <DebugMacros.h>
#include <DiagnosticMacros.h>
#include <BoopsiStubs.h>
#include <BoopsiLib.h>

#ifdef __STORM__
	#pragma header
#endif



/* Per object instance data */
struct ScrollButtonData
{
	/* The number of ticks we still have to wait
	 * before sending any notification.
	 */
	ULONG TickCounter;

	struct IBox GBox;		/* Our gadget bounds		*/
	struct IBox FrameBox;	/* Frame size and position	*/

	BOOL LayoutDone;		/* Need to recalculate above info? */
};



/* Local function prototypes */

static ULONG	SB_GMGoActive	(Class *cl, struct Gadget *g, struct gpInput *gpi);
static ULONG	SB_GMHandleInput(Class *cl, struct Gadget *g, struct gpInput *gpi);
static void		SB_GMRender		(Class *cl, struct Gadget *g, struct gpRender *gpr);
static void		SB_GMLayout		(Class *cl, struct Gadget *g, struct gpLayout *gpl);
static ULONG	SB_OMSet		(Class *cl, struct Gadget *g, struct opUpdate *msg);

static ULONG HOOKCALL ScrollButtonDispatcher(
	REG(a0, Class *cl),
	REG(a2, struct Gadget *g),
	REG(a1, Msg msg));



#if (CLASS_FLAVOUR & FLAVOUR_CLASSLIB)

	/* Class library support functions */
	struct ClassLibrary	*	HOOKCALL _UserLibInit		(REG(a6, struct ClassLibrary *mybase));
	struct ClassLibrary	*	HOOKCALL _UserLibCleanup	(REG(a6, struct ClassLibrary *mybase));
	Class *					HOOKCALL _GetEngine			(REG(a6, struct ClassLibrary *mybase));

	/* Library data */
	const UBYTE LibName[] = "scrollbutton.gadget";
	const UBYTE LibVer[] = { '$', 'V', 'E', 'R', ':', ' ' };
	const UBYTE LibId[] = "scrollbutton.gadget 1.0 (19.3.2000)  1997-2000 Bernardo Innocenti\n";

	/* Workaround a bug in StormC header file <proto/utility.h> */
	#ifdef __STORM__
		#define UTILITYBASETYPE struct Library
	#else
		#define UTILITYBASETYPE struct UtilityBase
	#endif

	/* Library bases */
	struct ExecBase			*SysBase		= NULL;
	struct IntuitionBase	*IntuitionBase	= NULL;
	UTILITYBASETYPE			*UtilityBase	= NULL;
#endif



static ULONG SB_GMGoActive(Class *cl, struct Gadget *g, struct gpInput *gpi)
{
	struct ScrollButtonData *bd = (struct ScrollButtonData *) INST_DATA(cl, g);

	DB2(DBPRINTF("ScrollButton: GM_GOACTIVE\n");)
	ASSERT_VALID_PTR(bd)

	/* May define an attribute to make delay configurable */
	bd->TickCounter = 3;

	/* Notify our target that we have initially hit. */
	NotifyAttrs((Object *)g, gpi->gpi_GInfo, 0,
		GA_ID,	g->GadgetID,
		TAG_DONE);

	/* Send more input */
	return GMR_MEACTIVE;
}



static ULONG SB_GMHandleInput(Class *cl, struct Gadget *g, struct gpInput *gpi)
{
	struct ScrollButtonData *bd = (struct ScrollButtonData *) INST_DATA(cl, g);
	struct RastPort *rp;
	ULONG retval = GMR_MEACTIVE;
	UWORD selected = 0;

	DB2 (DBPRINTF ("ScrollButton: GM_HANDLEINPUT\n");)
	ASSERT_VALID_PTR(bd)

	/* This also works with classic (non-boopsi) images. */
	if (PointInImage((gpi->gpi_Mouse.X << 16) + (gpi->gpi_Mouse.Y), g->GadgetRender))
	{
		/* We are hit */
		selected = GFLG_SELECTED;
	}

	if (gpi->gpi_IEvent->ie_Class == IECLASS_RAWMOUSE && gpi->gpi_IEvent->ie_Code == SELECTUP)
	{
		/* Gadgetup, time to go */
		retval = GMR_NOREUSE;
		/* Unselect the gadget on our way out... */
		selected = 0;
	}
	else if (gpi->gpi_IEvent->ie_Class == IECLASS_TIMER)
	{
		/* We got a tick. Decrement counter, and if 0, send notify. */

		if (bd->TickCounter)
			bd->TickCounter--;
		else if (selected)
			/* Notify our target that we are still being hit */
			NotifyAttrs ((Object *)g, gpi->gpi_GInfo, 0,
				GA_ID,	g->GadgetID,
				TAG_DONE);
	}

	if ((g->Flags & GFLG_SELECTED) != selected)
	{
		/* Update changes in gadget render */
		g->Flags ^= GFLG_SELECTED;
		if (rp = ObtainGIRPort (gpi->gpi_GInfo))
		{
			DoMethod((Object *) g, GM_RENDER, gpi->gpi_GInfo, rp, GREDRAW_UPDATE);
			ReleaseGIRPort(rp);
		}
	}
	return retval;
}



static void SB_GMRender(Class *cl, struct Gadget *g, struct gpRender *gpr)
{
	struct ScrollButtonData *bd = (struct ScrollButtonData *) INST_DATA(cl, g);
	struct impDraw impdraw;

	DB2(DBPRINTF("ScrollButton: GM_RENDER\n");)
	ASSERT_VALID_PTR(bd)

	if (!bd->LayoutDone)
		SB_GMLayout(cl, g, (struct gpLayout *)gpr);

	impdraw.MethodID	= IM_DRAWFRAME;
	impdraw.imp_RPort	= gpr->gpr_RPort;
	impdraw.imp_DrInfo	= gpr->gpr_GInfo->gi_DrInfo;

	/* Check if we are living inside an inactive window border */
	if ((g->Activation & (GACT_RIGHTBORDER | GACT_LEFTBORDER | GACT_TOPBORDER | GACT_BOTTOMBORDER)) &&
		gpr->gpr_GInfo->gi_Window &&
		(!(gpr->gpr_GInfo->gi_Window->Flags & WFLG_WINDOWACTIVE)))
	{
		impdraw.imp_State = (g->Flags & GFLG_SELECTED) ?
			IDS_INACTIVESELECTED :
			((g->Flags & GFLG_DISABLED) ? IDS_INACTIVEDISABLED : IDS_INACTIVENORMAL );
	}
	else
	{
		impdraw.imp_State = (g->Flags & GFLG_SELECTED) ?
			((g->Flags & GFLG_DISABLED) ? IDS_SELECTEDDISABLED : IDS_SELECTED ) :
			((g->Flags & GFLG_DISABLED) ? IDS_DISABLED : IDS_NORMAL );
	}

	/* Check whether our label image is valid and can be sent a boopsi message */
	if ((g->Flags & GFLG_LABELIMAGE) && g->GadgetText &&
		(((struct Image *)g->GadgetText)->Depth == CUSTOMIMAGEDEPTH))
	{
		impdraw.imp_Offset.X	= bd->GBox.Left - bd->FrameBox.Left;
		impdraw.imp_Offset.Y	= bd->GBox.Top - bd->FrameBox.Top;
		impdraw.imp_Dimensions.Width	= bd->GBox.Width - bd->FrameBox.Width;
		impdraw.imp_Dimensions.Height	= bd->GBox.Height - bd->FrameBox.Height;

		DoMethodA((Object *)g->GadgetText, (Msg)&impdraw);
	}

	/* Check whether our framing image is valid and can be sent a boopsi message */
	if ((g->Flags & GFLG_GADGIMAGE) && g->GadgetRender &&
		(((struct Image *)g->GadgetRender)->Depth == CUSTOMIMAGEDEPTH))
	{
		impdraw.imp_Offset.X	= bd->GBox.Left;
		impdraw.imp_Offset.Y	= bd->GBox.Top;
		impdraw.imp_Dimensions.Width	= bd->GBox.Width;
		impdraw.imp_Dimensions.Height	= bd->GBox.Height;

		DoMethodA((Object *)g->GadgetRender, (Msg)&impdraw);
	}
}



static void SB_GMLayout(Class *cl, struct Gadget *g, struct gpLayout *gpl)
{
	struct ScrollButtonData *bd = (struct ScrollButtonData *) INST_DATA(cl, g);

	DB2(DBPRINTF("ScrollButton: GM_LAYOUT\n");)
	ASSERT_VALID_PTR(bd)

	/* Compute new button size */
	GetGadgetBox(gpl->gpl_GInfo, g, &bd->GBox);

	/* Check if our framing image is valid and can be sent a boopsi message */
	if ((g->Flags & GFLG_GADGIMAGE) && g->GadgetRender &&
		(((struct Image *)g->GadgetRender)->Depth == CUSTOMIMAGEDEPTH))
	{
		struct IBox fakecontents = { 0, 0, 0, 0 };

		/* Ask the frame what is its size */
		DoMethod((Object *)g->GadgetRender, IM_FRAMEBOX,
			&fakecontents, &bd->FrameBox, gpl->gpl_GInfo->gi_DrInfo, 0);
	}

	bd->LayoutDone = TRUE;
}



static ULONG SB_OMSet (Class *cl, struct Gadget *g, struct opUpdate *msg)
{
	struct ScrollButtonData *bd = (struct ScrollButtonData *) INST_DATA(cl, g);
	struct TagItem	*ti, *tstate = msg->opu_AttrList;

	DB1(DBPRINTF("ScrollButton: OM_SET\n");)
	ASSERT_VALID_PTR(bd)
	ASSERT_VALID_PTR(tstate)

	while (ti = NextTagItem(&tstate))
	{
		/* Check for any of GA_Width, GA_RelWidth, GA_Height, GA_RelHeight.
		 * This test assumes that the values of these tags are consecutive
		 * and will break whenever their definitions are changed.
		 */
		if ((ti->ti_Tag >= GA_Width) && (ti->ti_Tag <= GA_RelHeight))
			bd->LayoutDone = FALSE;
	}

	/* The superclass handles everything else */
	return (DoSuperMethodA(cl, (Object *)g, (Msg) msg));
}



static ULONG HOOKCALL ScrollButtonDispatcher(
	REG(a0, Class *cl),
	REG(a2, struct Gadget *g),
	REG(a1, Msg msg))

/* ScrollButton Class Dispatcher entrypoint.
 * Handle boopsi messages.
 */
{
	switch (msg->MethodID)
	{
		case GM_GOACTIVE:
			return SB_GMGoActive(cl, g, (struct gpInput *)msg);

		case GM_HANDLEINPUT:
			return SB_GMHandleInput(cl, g, (struct gpInput *)msg);

		case GM_RENDER:
			SB_GMRender(cl, g, (struct gpRender *)msg);
			return TRUE;

		case GM_LAYOUT:
			SB_GMLayout(cl, g, (struct gpLayout *)msg);
			return TRUE;

		case OM_SET:
		case OM_UPDATE:
			return SB_OMSet(cl, g, (struct opUpdate *)msg);

		default:
			/* Super class handles everything else */
			return (DoSuperMethodA(cl, (Object *)g, msg));
	}
}



Class *MakeScrollButtonClass (void)
{
	Class *class;

	if (class = MakeClass(NULL, BUTTONGCLASS, NULL, sizeof(struct ScrollButtonData), 0))
		class->cl_Dispatcher.h_Entry = (ULONG (*)()) ScrollButtonDispatcher;

	return class;
}

Class *MakeFrScrollButtonClass (void)
{
	Class *class;

	if (class = MakeClass (NULL, FRBUTTONCLASS, NULL, sizeof(struct ScrollButtonData), 0))
		class->cl_Dispatcher.h_Entry = (ULONG (*)()) ScrollButtonDispatcher;

	return class;
}

BOOL FreeScrollButtonClass (Class *cl)
{
	return (FreeClass(cl));
}



#if (CLASS_FLAVOUR & FLAVOUR_CLASSLIB)
/*
 * Class library support functions
 */

struct ClassLibrary * HOOKCALL _UserLibInit(REG(a6, struct ClassLibrary *mybase))
{
	/* Initialize SysBase */
	SysBase = *((struct ExecBase **)4UL);

	if (!((UtilityBase = (UTILITYBASETYPE *) OpenLibrary("utility.library", 39)) &&
		(IntuitionBase = (struct IntuitionBase *)OpenLibrary("intuition.library", 39))))
	{
		_UserLibCleanup(mybase);
		return NULL;
	}

	if (!(mybase->cl_Class = MakeScrollButtonClass()))
	{
		_UserLibCleanup(mybase);
		return NULL;
	}

	AddClass(mybase->cl_Class);
	return mybase;
}

struct ClassLibrary * HOOKCALL _UserLibCleanup(REG(a6, struct ClassLibrary *mybase))
{
	if (mybase->cl_Class)
		if (!FreeScrollButtonClass(mybase->cl_Class))
			return NULL;

	CloseLibrary((struct Library *)IntuitionBase);
	CloseLibrary((struct Library *)UtilityBase);

	return mybase;
}

Class * HOOKCALL _GetEngine(REG(a6, struct ClassLibrary *mybase))
{
	return mybase->cl_Class;
}

#endif /* (CLASS_FLAVOUR & FLAVOUR_CLASSLIB) */
