/*
**	$Id: FrPropGClass.c,v 1.1 1999/02/07 14:24:42 bernie Exp $
**
**	Copyright (C) 1999,2000 Bernardo Innocenti
**	All rights reserved.
**
**	Use 4 chars wide TABs to read this file
**
**	A simple subclass of the propgclass which can be surrounded
**	by a boopsi frame image specified through the GA_LabelImage tag.
*/

#define USE_BUILTIN_MATH
#define INTUI_V36_NAMES_ONLY
#define __USE_SYSBASE
#define  CLIB_ALIB_PROTOS_H		/* Avoid dupe defs of boopsi funcs */


#include <intuition/gadgetclass.h>
#include <intuition/imageclass.h>
#include <proto/intuition.h>

#ifdef __STORM__
	#pragma header
#endif

#include "CompilerSpecific.h"
#include "DebugMacros.h"
#include "BoopsiStubs.h"

#include "FrPropGClass.h"




/* Per object instance data */
struct FrPropData
{
	Object *	Frame;
	struct IBox FrameBox;
};



/* ScrollButton Class Dispatcher entrypoint.
 * Handle boopsi messages.
 */
static ULONG HOOKCALL FrPropDispatcher (
	REG(a0, Class * cl),
	REG(a2, struct Gadget * g),
	REG(a1, Msg msg))
{
	struct FrPropData *fpd;

	switch (msg->MethodID)
	{
		case OM_NEW:
			DB2(DBPRINTF("FrPropGClass: OM_NEW\n");)

			if (g = (struct Gadget *)DoSuperMethodA (cl, (Object *)g, (Msg)msg))
			{
				fpd = (struct FrPropData *) INST_DATA(cl, g);

				/* HACK: the gadgetclass copies the contents
				 * of the GA_LabelImage tag into the Gadget structure.
				 * We steal the frame from there and then clear it,
				 * otherwise the gadgetclass will try to draw the frame
				 * itself.
				 */
				fpd->Frame = (Object *) g->GadgetText;
				g->GadgetText = NULL;
				g->Flags &= ~GFLG_LABELIMAGE;

				/* Ask Intuition to send us GM_LAYOUT messages */
				g->Flags |= GFLG_RELSPECIAL;
			}
			return (ULONG)g;

		case GM_LAYOUT:
			DB2(DBPRINTF("FrPropGClass: GM_LAYOUT\n");)

			fpd = (struct FrPropData *) INST_DATA(cl, g);

			if (fpd->Frame)
			{
				struct IBox gbox;

				/* Compute current gadget size */
				GetGadgetBox(((struct gpLayout *)msg)->gpl_GInfo, g, &gbox);

				DB2(DBPRINTF("FrPropGClass:   GBox :%ld t:%ld w:%ld h:%ld\n",
						fpd->FrameBox.Left, fpd->FrameBox.Top,
						fpd->FrameBox.Width, fpd->FrameBox.Height);)

				/* Calculate dimensions of our framing image */
				DoMethod(fpd->Frame, IM_FRAMEBOX, &gbox, &fpd->FrameBox,
					((struct gpLayout *)msg)->gpl_GInfo->gi_DrInfo, 0);
			}
			break; /* go to DoSuperMethod() below */

		case GM_RENDER:
			if (((struct gpRender *)msg)->gpr_Redraw == GREDRAW_REDRAW)
			{
				fpd = (struct FrPropData *) INST_DATA(cl, g);
				DB2(DBPRINTF("FrPropGClass: GM_RENDER - GREDRAW_REDRAW\n");)
				ASSERT_VALID_PTR(fpd)

				if (fpd->Frame)
				{
					ASSERT_VALID_PTR(fpd->Frame)
					DB2(DBPRINTF("FrPropGClass:   Frame l:%ld t:%ld w:%ld h:%ld\n",
						fpd->FrameBox.Left, fpd->FrameBox.Top,
						fpd->FrameBox.Width, fpd->FrameBox.Height);)

					DoMethod(fpd->Frame, IM_DRAWFRAME,
						((struct gpRender *)msg)->gpr_RPort,					/* imp_RPort		*/
						(fpd->FrameBox.Left << 16) | (fpd->FrameBox.Top),		/* imp_Offset		*/
						IDS_NORMAL,												/* imp_State		*/
						((struct gpRender *)msg)->gpr_GInfo->gi_DrInfo,			/* imp_DrInfo		*/
						(fpd->FrameBox.Width << 16) | (fpd->FrameBox.Height));	/* imp_Dimensions	*/
				}
			}
			break; /* go to DoSuperMethod() below */

		case OM_UPDATE:
		{
			ULONG result;
			Class *oldclass;

			/* This is a dirty workaround for a bug in the propgclass.
			 *
			 * It appears that the dispatcher of the propgclass detects
			 * when an object is made from a subclass. In this case,
			 * it stops sending itself the GM_REDRAW message when the
			 * PGA_#? attributes are updated. The result is that the
			 * prop does not update its visual when the attributes are
			 * changed.
			 *
			 * The workaround we use here is to temporarly switch the
			 * object true class to propgclass before passing the
			 * OM_UPDATE method to our superclass.
			 */
			oldclass = OCLASS(g);
			OCLASS(g) = cl->cl_Super;
			result = CoerceMethodA (cl->cl_Super, (Object *)g, msg);
			OCLASS(g) = oldclass;
			return result;
		}

		default:
			break; /* go to DoSuperMethod() below */
	}

	/* Super class handles everything */
	return (DoSuperMethodA (cl, (Object *)g, msg));
}



Class * MakeFrPropGClass(void)
{
	Class * class;

	if (class = MakeClass (NULL, PROPGCLASS, NULL, sizeof(struct FrPropData), 0))
		class->cl_Dispatcher.h_Entry = (ULONG (*)()) FrPropDispatcher;

	return class;
}



BOOL FreeFrPropGClass(Class *cl)
{
	return (FreeClass(cl));
}
