/*
**	$Id: SliderBarDemo.c,v 1.1 1999/02/07 14:24:43 bernie Exp $
**
**	Copyright (C) 1999 Bernardo Innocenti (<bernardo.innocenti@usa.net>)
**	All rights reserved.
**
**	Use 4 chars wide TABs to read this file
**
**
**	Introduction
**	============
**
**	This program shows the usage of the `boopsi' "sliderbargclass" gadget.
**	It can only be started from the shell (not from Workbench).
**
**
**	Compiling
**	=========
**
**	This project can be compiled with SAS/C 6.58 or better and
**	StormC 2.00.23 or better. You get the smaller and faster
**	executable with SAS/C. SAS/C will give you a couple of warnings
**	on the library bases which are defined static. This warning can't be
**	fixed without patching the system headers. Don't worry about it.
**
**
**	History
**	=======
**
**	1.0 (30.1.99)	First release
**
**
**	Known Bugs
**	==========
**
**		- This code has never been tested on V37.
**
**
**	License
**	=======
**
**	This software is freely distributable as long as the source code,
**	documentation and executables are kept together.  Permission is granted
**	to release modified versions of this program as long as all existing
**	copyright notices are left intact.
*/

#define INTUI_V36_NAMES_ONLY
#define __USE_SYSBASE
#define  CLIB_ALIB_PROTOS_H	/* Avoid including this header file because of
							 * conflicting definitions in BoopsiStubs.h
							 */

#include <exec/types.h>
#include <exec/memory.h>
#include <exec/execbase.h>

#include <dos/dos.h>

#include <intuition/intuition.h>
#include <intuition/intuitionbase.h>
#include <intuition/screens.h>
#include <intuition/classes.h>
#include <intuition/gadgetclass.h>
#include <intuition/icclass.h>

#include <proto/exec.h>
#include <proto/intuition.h>

#include "CompilerSpecific.h"
#include "DebugMacros.h"
#include "BoopsiStubs.h"

#include "SliderBarGClass.h"


/* Local function prototypes */
LONG SAVEDS					 main			(void);
static struct SBDHandle		*OpenSBDWin		(UBYTE *pubscreen, ULONG left, ULONG top,
											ULONG width, ULONG height);
static void					 CloseSBDWin	(struct SBDHandle *sbdhandle);
static struct Gadget		*CreateGadgets	(struct SBDHandle *sbdhandle);
static void					 DisposeGadgets (struct SBDHandle *sbdhandle);
static struct ClassLibrary	*OpenClass		(STRPTR name, ULONG version);



/* This structure describes an open SliderBarDemo window */

struct SBDHandle
{
	struct Window	*Win;
	struct Screen	*Scr;
	struct Gadget	*Slider[4];
};



/* Version tag */

static UBYTE versiontag[] = "$VER: SlideBarDemo 1.0 (30.1.99) by Bernardo Innocenti";


/* Library bases */
static struct ExecBase		*SysBase;
static struct IntuitionBase	*IntuitionBase;

/* `boopsi' class library base */
static struct ClassLibrary	*SliderBarBase;



LONG SAVEDS _main (void)

/* Main program entry point.  When linking without startup code, this
 * must be the first function in the first object module listed on the
 * linker command line.  We also need to initialize SysBase and open
 * all needed libraries manually.
 */
{
	struct SBDHandle *sbdhandle;
	LONG			 sigwait, sigrcvd;
	LONG			 retval	= RETURN_FAIL;
	BOOL			 quit	= FALSE;


	/* Initialize SysBase */
	SysBase = *((struct ExecBase **)4UL);

	if (IntuitionBase = (struct IntuitionBase *)OpenLibrary ("intuition.library", 39L))
	{
		if (SliderBarBase = OpenClass ("sliderbar.gadget", 0))
		{
			if (sbdhandle = OpenSBDWin (NULL, 0, 20, 320, 200))
			{
				/* Pre-calculate the signal mask for Wait() */
				sigwait = (1 << sbdhandle->Win->UserPort->mp_SigBit) |
					SIGBREAKF_CTRL_C;

				/* Now for the main loop.  As you can see, it is really
				 * very compact. That's the magic of boopsi! :-)
				 */
				while (!quit)
				{
					/* Sleep until something interesting occurs */
					sigrcvd = Wait (sigwait);

					/* Now handle received signals */

					/* Break signal? */
					if (sigrcvd & SIGBREAKF_CTRL_C)
						quit = TRUE;

					/* IDCMP message? */
					if (sigrcvd & (1 << sbdhandle->Win->UserPort->mp_SigBit))
					{
						struct IntuiMessage	*msg;

						while (msg = (struct IntuiMessage *) GetMsg (sbdhandle->Win->UserPort))
						{
							switch (msg->Class)
							{
								case IDCMP_CLOSEWINDOW:
									quit = TRUE;
									break;

								default:
									break;
							}
							ReplyMsg ((struct Message *) msg);
						}
					}
				} /* End while (!quit) */

				retval = RETURN_OK;

				CloseSBDWin (sbdhandle);
			}

			/* This cannot fail. Passing NULL is ok. */
			CloseLibrary ((struct Library *)SliderBarBase);
		}
		CloseLibrary ((struct Library *)IntuitionBase);
	}

	return retval;
}



static struct SBDHandle *OpenSBDWin (UBYTE *pubscreen,
	ULONG left, ULONG top, ULONG width, ULONG height)
{
	struct SBDHandle	*sbdhandle;
	struct Gadget		*glist;

	if (sbdhandle = AllocMem (sizeof (struct SBDHandle), MEMF_ANY | MEMF_CLEAR))
	{
		if (sbdhandle->Scr = LockPubScreen (pubscreen))
		{
			if (glist = CreateGadgets (sbdhandle))
			{
				if (sbdhandle->Win = OpenWindowTags (NULL,
					WA_Top,				top,
					WA_Left,			left,
					WA_InnerWidth,		width,
					WA_InnerHeight,		height,
					WA_PubScreen,		sbdhandle->Scr,
					WA_PubScreenFallBack, TRUE,
					WA_AutoAdjust,		TRUE,
					WA_IDCMP,			IDCMP_CLOSEWINDOW,
					WA_Flags,			WFLG_DRAGBAR | WFLG_DEPTHGADGET | WFLG_CLOSEGADGET |
										WFLG_SIZEGADGET | WFLG_SIMPLE_REFRESH | WFLG_NOCAREREFRESH,
					WA_Gadgets,			glist,
					WA_Title,			versiontag + 6,
					WA_MinWidth,		64,
					WA_MinHeight,		64,
					WA_MaxWidth,		-1,
					WA_MaxHeight,		-1,
					TAG_DONE))
				{
					UnlockPubScreen (NULL, sbdhandle->Scr);
					return sbdhandle;
				}
				DisposeGadgets (sbdhandle);
			}
			UnlockPubScreen (NULL, sbdhandle->Scr);
		}
		FreeMem (sbdhandle, sizeof (struct SBDHandle));
	}
	return NULL;
}



static void CloseSBDWin (struct SBDHandle *sbdhandle)
{
	/* Close our window. No need to reply queued messages,
	 * Intuition is clever enough to care about this for us.
	 */
	CloseWindow (sbdhandle->Win);
	DisposeGadgets (sbdhandle);
	FreeMem (sbdhandle, sizeof (struct SBDHandle));
}

#define SLIDER_THICKNESS 20

static struct Gadget *CreateGadgets (struct SBDHandle *sbdhandle)
{
	struct Screen	*scr = sbdhandle->Scr;
	LONG left,right, top, bottom, width, height;


	left	= scr->WBorLeft + 10;
	right	= - scr->WBorRight - 25 - SLIDER_THICKNESS;
	top		= scr->WBorTop + scr->Font->ta_YSize + 10;
	bottom	= - scr->WBorBottom - 10 - SLIDER_THICKNESS;
	width	= right - left - 1;
	height	= bottom - top - 1;

	sbdhandle->Slider[0] = NewObject (NULL, SLIDERBARGCLASS,
		GA_Left,		left,
		GA_Top,			top,
		GA_Width,		SLIDER_THICKNESS,
		GA_RelHeight,	height,
		PGA_Total,		120,
		PGA_Top,		50,
		PGA_Visible,	20,
		LAYOUTA_Orientation, LORIENT_VERT,
		TAG_DONE);

	sbdhandle->Slider[1] = NewObject (NULL, SLIDERBARGCLASS,
		GA_Previous,	sbdhandle->Slider[0],
		GA_Left,		left,
		GA_RelBottom,	bottom,
		GA_RelWidth,	width + SLIDER_THICKNESS,
		GA_Height,		SLIDER_THICKNESS,
		PGA_Total,		140,
		PGA_Top,		50,
		PGA_Visible,	40,
		LAYOUTA_Orientation, LORIENT_HORIZ,
		ICA_TARGET,		sbdhandle->Slider[0],
		TAG_DONE);

	sbdhandle->Slider[2] = NewObject (NULL, SLIDERBARGCLASS,
		GA_Previous,	sbdhandle->Slider[1],
		GA_RelRight,	right,
		GA_Top,			top,
		GA_Width,		SLIDER_THICKNESS,
		GA_RelHeight,	height,
		PGA_Total,		140,
		PGA_Top,		50,
		PGA_Visible,	40,
		LAYOUTA_Orientation, LORIENT_VERT,
		ICA_TARGET,		sbdhandle->Slider[1],
		TAG_DONE);

	sbdhandle->Slider[3] = NewObject (NULL, SLIDERBARGCLASS,
		GA_Previous,	sbdhandle->Slider[2],
		GA_Top,			top,
		GA_Left,		left + SLIDER_THICKNESS,
		GA_RelWidth,	width - SLIDER_THICKNESS,
		GA_Height,		SLIDER_THICKNESS,
		PGA_Total,		140,
		PGA_Top,		50,
		PGA_Visible,	40,
		LAYOUTA_Orientation, LORIENT_HORIZ,
		ICA_TARGET,		sbdhandle->Slider[2],
		TAG_DONE);


	if (sbdhandle->Slider[0] &&
		sbdhandle->Slider[1] &&
		sbdhandle->Slider[2] &&
		sbdhandle->Slider[3])
	{
		return sbdhandle->Slider[0];
	}
	else
	{
		DisposeGadgets (sbdhandle);
		return NULL;
	}
}



static void DisposeGadgets (struct SBDHandle *sbdhandle)
{
	int i;

	for (i = 0; i < 4; i++)
		DisposeObject (sbdhandle->Slider[i]);
}



static struct ClassLibrary *OpenClass (STRPTR name, ULONG version)

/* Open named class. Look both in current and gadgets/ directory
 */
{
	static struct EasyStruct OpenClassES =
	{
		sizeof (struct EasyStruct),
		0,
		versiontag + 6,
		"Couldn't open %s version %ld or greater",
		"Ok"
	};
	static const char prefix[] = "gadgets/";

	struct ClassLibrary *classbase;
	char buf[256];
	int i;


	if (!(classbase = (struct ClassLibrary *)OpenLibrary (name, version)))
	{
		/* We can't use AddPart() here because we didn't open dos.library */

		/* Copy the prefix in the buffer */
		for (i = 0; buf[i] = prefix[i]; i++);

		/* Append the name */
		while (buf[i++] = *name++);

		/* Try again */
		classbase = (struct ClassLibrary *)OpenLibrary (buf, version);

		if (!classbase)
		{
			/* Report an error */
			EasyRequest (NULL, &OpenClassES, NULL, buf, version);
		}
	}
	return classbase;
}
