/*
**	$Id: SliderBarGClass.c,v 1.1 1999/02/07 14:24:43 bernie Exp $
**
**	Copyright (C) 1999 Bernardo Innocenti
**	All rights reserved.
**
**	Use 4 chars wide TABs to read this file
**
**	"sliderbargclass", a GadTools like slider with scroll buttons.
**	See the header file "SliderBarGClass.h" for class documentation.
*/

/* Supported class flavours:
 *
 * FLAVOUR_CLASSLIB
 *	Build a class library if this option is set or
 *	a static linker object otherwise.
 *
 * FLAVOUR_PUBCLASS
 *	Call AddClass() to show this class globally if set,
 *	make a private class otherwise.
 */
#define FLAVOUR_CLASSLIB	(1<<0)
#define FLAVOUR_PUBCLASS	(1<<1)

#define CLASS_FLAVOUR (FLAVOUR_CLASSLIB | FLAVOUR_PUBCLASS)

/* Definitions for system headers */
#define USE_BUILTIN_MATH
#define INTUI_V36_NAMES_ONLY
#define INTUITION_IOBSOLETE_H
#define __USE_SYSBASE
#define  CLIB_ALIB_PROTOS_H		/* Avoid dupe defines of boopsi funcs */

#include <exec/types.h>
#include <exec/memory.h>
#include <utility/tagitem.h>
#include <intuition/intuition.h>
#include <intuition/classes.h>
#include <intuition/classusr.h>
#include <intuition/gadgetclass.h>
#include <intuition/imageclass.h>
#include <intuition/icclass.h>

#include <proto/exec.h>
#include <proto/intuition.h>
#include <proto/utility.h>

#include <CompilerSpecific.h>
#include <DebugMacros.h>
#include <DiagnosticMacros.h>
#include <BoopsiStubs.h>

#include <images/VectorGlyphIClass.h>
#include <gadgets/SmartGroupGClass.h>
#include <gadgets/ScrollButtonClass.h>
#include "FrPropGClass.h"
#include "SliderBarGClass.h"

#ifdef __STORM__
	#pragma header
#endif


#define	BUTTON_WIDTH	16
#define	BUTTON_HEIGHT	16

/* Local function prototypes */

INLINE void GetGadgetBox (struct GadgetInfo *ginfo, struct Gadget *g, struct IBox *box);

static ULONG HOOKCALL SliderBarDispatcher (
	REG(a0, Class *cl),
	REG(a2, struct Gadget *g),
	REG(a1, Msg msg));

static void		SB_GMLayout	(Class *cl, struct Gadget *g, struct gpLayout *msg);
static ULONG	SB_OMNew	(Class *cl, struct Gadget *g, struct opSet *msg);
static void		SB_OMDispose(Class *cl, struct Gadget *g, Msg msg);
static ULONG	SB_OMGet	(Class *cl, struct Gadget *g, struct opGet *msg);
static ULONG	SB_OMSet	(Class *cl, struct Gadget *g, struct opUpdate *msg);



/* Per-object instance data */
struct SBData
{
	/* Children */
	Object *Prop, *DecButton, *IncButton;

	/* VectorGlyph images for the buttons */
	Object *IncImage, *DecImage;

	/* Copy of prop attributes */
	LONG	Top, Total, Visible;

	/* Slider orientation: can be LORIENT_VERT or LORIENT_HORIZ */
	ULONG	Orientation;
};



/* Global class data */
struct SBClassData
{
	/* Boopsi frame to put around our children */
	Object				*Frame;

	/* Classes we are using */
	Class				*SmartGroupGClass;
	Class				*FrScrollButtonClass;
	Class				*FrPropGClass;
	Class				*VectorGlyphIClass;
	struct ClassLibrary	*VectorGlyphBase;
};


#if (CLASS_FLAVOUR & FLAVOUR_CLASSLIB)

	/* Class library support functions */
	struct ClassLibrary	*	HOOKCALL _UserLibInit		(REG(a6, struct ClassLibrary *mybase));
	struct ClassLibrary	*	HOOKCALL _UserLibCleanup	(REG(a6, struct ClassLibrary *mybase));
	Class *					HOOKCALL _GetEngine			(REG(a6, struct ClassLibrary *mybase));

	/* Library data */
	const UBYTE LibName[] = "sliderbar.gadget";
	const UBYTE LibVer[] = { '$', 'V', 'E', 'R', ':', ' ' };
	const UBYTE LibId[] = "sliderbar.gadget 1.0 (30.1.99)  1999 Bernardo Innocenti\n";

	/* Workaround a bug in StormC header file <proto/utility.h> */
	#ifdef __STORM__
		#define UTILITYBASETYPE struct Library
	#else
		#define UTILITYBASETYPE struct UtilityBase
	#endif

	/* Library bases - also used by the FrScrollButtonClass and the FrPropGClass */
	struct ExecBase			*SysBase		= NULL;
	struct IntuitionBase	*IntuitionBase	= NULL;
	UTILITYBASETYPE			*UtilityBase	= NULL;
#endif


/* Static variables */

static LONG MapPropToSB[] =
{
	PGA_Top,	SBA_InternalTop,
	TAG_DONE
};


static LONG MapButtonToSB[] =
{
	GA_ID,		SBA_Adjust,
	TAG_DONE
};



INLINE void GetGadgetBox (struct GadgetInfo *ginfo, struct Gadget *g, struct IBox *box)

/* Gets the actual IBox where a gadget exists in a window.
 * The special cases it handles are all the REL#? (relative positioning flags).
 *
 * This function returns the gadget size in the provided IBox structure,
 * computing the values from the coordinates of the gadget and the window
 * inside which it lives.
 */
{
	ASSERT_VALID_PTR(g)
	ASSERT_VALID_PTR(ginfo)
	ASSERT_VALID_PTR(box)

	box->Left = g->LeftEdge;
	if (g->Flags & GFLG_RELRIGHT)
		box->Left += ginfo->gi_Domain.Width - 1;

	box->Top = g->TopEdge;
	if (g->Flags & GFLG_RELBOTTOM)
		box->Top += ginfo->gi_Domain.Height - 1;

	box->Width = g->Width;
	if (g->Flags & GFLG_RELWIDTH)
		box->Width += ginfo->gi_Domain.Width;

	box->Height = g->Height;
	if (g->Flags & GFLG_RELHEIGHT)
		box->Height += ginfo->gi_Domain.Height;
}



static void SB_GMLayout (Class *cl, struct Gadget *g, struct gpLayout *msg)
{
	struct SBData		*sb = (struct SBData *) INST_DATA (cl, (Object *)g);
	struct SBClassData	*classdata = (struct SBClassData *)cl->cl_UserData;
	struct IBox gbox, framebox, contentsbox = { 0, 0, 0, 0 };

	DB2 (kprintf ("SliderBarClass: GM_LAYOUT\n");)
	ASSERT_VALID_PTR(sb)
	ASSERT_VALID_PTR(classdata)


	/* Collect new group size */
	GetGadgetBox (msg->gpl_GInfo, g, &gbox);

	/* Query the frame on its nominal width and height */
	DoMethod (classdata->Frame, IM_FRAMEBOX, &contentsbox, &framebox,
		msg->gpl_GInfo->gi_DrInfo, 0);

	/* Size our children accordingly */

	if (sb->Orientation == LORIENT_VERT)
	{
		SetAttrs (sb->Prop,
			GA_Left,	gbox.Left + framebox.Width / 2,
			GA_Top,		gbox.Top + framebox.Height / 2,
			GA_Width,	gbox.Width - framebox.Width,
			GA_Height,	gbox.Height - framebox.Height - BUTTON_HEIGHT * 2,
			TAG_DONE);

		if (sb->DecButton)
			SetAttrs (sb->DecButton,
				GA_Left,	gbox.Left,
				GA_Top,		gbox.Top + gbox.Height - BUTTON_HEIGHT * 2,
				GA_Width,	gbox.Width,
				GA_Height,	BUTTON_HEIGHT,
				TAG_DONE);

		if (sb->IncButton)
			SetAttrs (sb->IncButton,
				GA_Left,	gbox.Left,
				GA_Top,		gbox.Top + gbox.Height - BUTTON_HEIGHT,
				GA_Width,	gbox.Width,
				GA_Height,	BUTTON_HEIGHT,
				TAG_DONE);
	}
	else /* LORIENT_HORIZ */
	{
		SetAttrs (sb->Prop,
			GA_Left,	gbox.Left + framebox.Width / 2,
			GA_Top,		gbox.Top + framebox.Height / 2,
			GA_Width,	gbox.Width - framebox.Width - BUTTON_WIDTH * 2,
			GA_Height,	gbox.Height - framebox.Height,
			TAG_DONE);

		if (sb->DecButton)
			SetAttrs (sb->DecButton,
				GA_Left,	gbox.Left + gbox.Width - BUTTON_WIDTH * 2,
				GA_Top,		gbox.Top,
				GA_Width,	BUTTON_WIDTH,
				GA_Height,	gbox.Height,
				TAG_DONE);

		if (sb->IncButton)
			SetAttrs (sb->IncButton,
				GA_Left,	gbox.Left + gbox.Width - BUTTON_WIDTH,
				GA_Top,		gbox.Top,
				GA_Width,	BUTTON_WIDTH,
				GA_Height,	gbox.Height,
				TAG_DONE);
	}

	DoSuperMethodA (cl, (Object *)g, (Msg)msg);
}



static ULONG SB_OMNew (Class *cl, struct Gadget *g, struct opSet *msg)
{
	struct SBData		*sb;
	struct SBClassData	*classdata;

	DB2 (DBPRINTF ("SliderBarGClass: OM_NEW\n");)

	if (g = (struct Gadget *)DoSuperMethodA (cl, (Object *)g, (Msg)msg))
	{
		classdata = (struct SBClassData *)cl->cl_UserData;
		sb = (struct SBData *) INST_DATA (cl, (Object *)g);
		ASSERT_VALID_PTR(sb)
		ASSERT_VALID_PTR(classdata)


		/* Initialize the object instance */
		sb->DecButton = NULL;
		sb->IncButton = NULL;
		sb->IncImage = NULL;
		sb->DecImage = NULL;
		sb->Orientation = GetTagData (LAYOUTA_Orientation, LORIENT_HORIZ, msg->ops_AttrList);
		sb->Top		= GetTagData (PGA_Top, 0, msg->ops_AttrList);
		sb->Total	= GetTagData (PGA_Total, 0, msg->ops_AttrList);
		sb->Visible	= GetTagData (PGA_Visible, 0, msg->ops_AttrList);

		if (sb->Prop = NewObject (classdata->FrPropGClass, NULL,
			GA_ID,			g->GadgetID,	/* Same as our ID */
			GA_LabelImage,	classdata->Frame,
			PGA_Freedom,	(sb->Orientation == LORIENT_VERT) ? FREEVERT : FREEHORIZ,
			PGA_NewLook,	TRUE,
			PGA_Borderless,	TRUE,
			PGA_Top,		sb->Top,
			PGA_Total,		sb->Total,
			PGA_Visible,	sb->Visible,
			ICA_TARGET,		g,	/* target on us */
			ICA_MAP,		MapPropToSB,
			TAG_DONE))
		{
			/* From now on, the groupgclass will dispose this object for us */
			DoMethod ((Object *)g, OM_ADDMEMBER, sb->Prop);

			/* Don't bother checking for failure here:
			 * the images are not mission-critical.
			 */
			sb->DecImage = NewObject (classdata->VectorGlyphIClass, NULL,
				SYSIA_Which, (sb->Orientation == LORIENT_VERT) ? VG_UPARROW : VG_LEFTARROW,
				TAG_DONE);

			if (sb->DecButton = NewObject (classdata->FrScrollButtonClass, NULL,
				GA_ID,			-1,
				GA_LabelImage,	sb->DecImage,
				GA_Image,		classdata->Frame,
				ICA_TARGET,		g, /* target on us */
				ICA_MAP,		MapButtonToSB,
				TAG_DONE))
			{
				/* From now on, the groupgclass will dispose this object for us */
				DoMethod ((Object *)g, OM_ADDMEMBER, sb->DecButton);
			}

			/* Don't bother checking for failure here:
			 * the images are not mission-critical.
			 */
			sb->IncImage = NewObject (classdata->VectorGlyphIClass, NULL,
				SYSIA_Which, (sb->Orientation == LORIENT_VERT) ? VG_DOWNARROW : VG_RIGHTARROW,
				TAG_DONE);

			if (sb->IncButton = NewObject (classdata->FrScrollButtonClass, NULL,
				GA_ID,			+1,
				GA_LabelImage,	sb->IncImage,
				GA_Image,		classdata->Frame,
				ICA_TARGET,		g, /* target on us */
				ICA_MAP,		MapButtonToSB,
				TAG_DONE))
			{
				/* From now on, the groupgclass will dispose this object for us */
				DoMethod ((Object *)g, OM_ADDMEMBER, sb->IncButton);
			}

			return (ULONG)g;	/* Return newly created instance */
		}

		/* Commit suicide without disturbing the dispatchers of our subclasses */
		CoerceMethod (cl, (Object *)g, OM_DISPOSE);
	}

	return 0;	/* Fail */
}



static void SB_OMDispose (Class *cl, struct Gadget *g, Msg msg)
{
	struct SBData *sb = (struct SBData *) INST_DATA (cl, (Object *)g);

	/* The groupgclass will dispose all our children for us */

	DisposeObject (sb->IncImage);
	DisposeObject (sb->DecImage);

	/* Our superclass will cleanup everything else now */
	DoSuperMethodA (cl, (Object *)g, msg);

	/* From now on, our instance data is no longer available */
}


static ULONG SB_OMSet (Class *cl, struct Gadget *g, struct opUpdate *msg)
{
	struct SBData	*sb = INST_DATA (cl, g);
	struct TagItem	*ti, *tstate = msg->opu_AttrList;
	ULONG action = 0;	/* see flag definitions below */
	ULONG result;

	ASSERT_VALID_PTR(sb)
	ASSERT_VALID_PTR_OR_NULL(tstate)

	DB2 (DBPRINTF ((msg->MethodID == OM_SET) ?
		"SliderBarGClass: OM_SET\n" :
		"SliderBarGClass: OM_UPDATE\n");)


	/* Definitions for the ations to be taken right after
	 * scanning the attributes list in OM_SET/OM_UPDATE.
	 * For speed reasons we pack them together in a single variable,
	 * so we can set and test multiple flags in once.
	 */
	#define ACTIONF_DO_SUPER_METHOD	(1<<0)
	#define ACTIONF_NOTIFY_TOP		(1<<1)
	#define ACTIONF_NOTIFY_ALL		(1<<2)
	#define ACTIONF_UPDATE_CHILD	(1<<3)


	while (ti = NextTagItem (&tstate))
		switch (ti->ti_Tag)
		{
			case GA_ID:
				if (msg->MethodID == OM_SET)
				{
					DB2 (DBPRINTF ("  GA_ID, %ld\n", ti->ti_Data);)

					/* Avoid forwarding all taglists to superclass because of GA_ID */
					g->GadgetID = ti->ti_Data;
				}
				break;

			case PGA_Top:
				if (sb->Top != ti->ti_Data)
				{
					sb->Top = ti->ti_Data;
					action |= ACTIONF_NOTIFY_TOP | ACTIONF_UPDATE_CHILD;
				}
				break;

			case PGA_Total:
				if (sb->Total != ti->ti_Data)
				{
					sb->Total = ti->ti_Data;
					action |= ACTIONF_NOTIFY_ALL | ACTIONF_UPDATE_CHILD;
				}
				break;

			case PGA_Visible:
				if (sb->Visible != ti->ti_Data)
				{
					sb->Visible = ti->ti_Data;
					action |= ACTIONF_NOTIFY_ALL | ACTIONF_UPDATE_CHILD;
				}
				break;

			case SBA_InternalTop:
				/* Prop gadget updating us */
				if (sb->Top != ti->ti_Data)
				{
					sb->Top = ti->ti_Data;
					action |= ACTIONF_NOTIFY_TOP;
				}
				break;

			case SBA_Adjust:
			{
				LONG newtop = sb->Top + (LONG)((WORD)ti->ti_Data);

				/* Clip result to allowable values */
				if (newtop < 0) newtop = 0;
				if (newtop > sb->Total - sb->Visible) newtop = sb->Total - sb->Visible;

				if (newtop != sb->Top)
				{
					sb->Top = newtop;
					action |= ACTIONF_NOTIFY_TOP | ACTIONF_UPDATE_CHILD;
				}

				break;
			}

			default:
				action |= ACTIONF_DO_SUPER_METHOD;
				break;
		}

	/* Forward method to our superclass dispatcher, only if needed */
	if (action & ACTIONF_DO_SUPER_METHOD)
	{
		DB2 (DBPRINTF ("SliderBarGClass:   Forwarding OM_SET/OM_UPDATE to superclass\n"));
		result = (DoSuperMethodA (cl, (Object *)g, (Msg) msg));
	}
	else
		result = TRUE;

	if (action & ACTIONF_UPDATE_CHILD)
	{
		if (action & ACTIONF_NOTIFY_ALL)
		{
			DB2 (DBPRINTF ("SliderBarGClass:   Update child ALL\n"));

			UpdateAttrs (sb->Prop, msg->opu_GInfo,
				(msg->MethodID == OM_UPDATE) ? msg->opu_Flags : 0,
				PGA_Top,		sb->Top,
				PGA_Total,		sb->Total,
				PGA_Visible,	sb->Visible,
				TAG_DONE);
		}
		else /* Top only */
		{
			DB2 (DBPRINTF ("SliderBarGClass:   Update child PGA_Top (%ld)\n", sb->Top));

			UpdateAttrs (sb->Prop, msg->opu_GInfo,
				(msg->MethodID == OM_UPDATE) ? msg->opu_Flags : 0,
				PGA_Top,		sb->Top,
				TAG_DONE);
		}
	}

	if (action & ACTIONF_NOTIFY_ALL)
	{
		DB2 (DBPRINTF ("SliderBarGClass:   Notify target ALL\n"));

		NotifyAttrs ((Object *)g, msg->opu_GInfo,
			(msg->MethodID == OM_UPDATE) ? msg->opu_Flags : 0,
			PGA_Top,		sb->Top,
			PGA_Total,		sb->Total,
			PGA_Visible,	sb->Visible,
			GA_ID,			g->GadgetID,
			TAG_DONE);
	}
	else if (action & ACTIONF_NOTIFY_TOP)
	{
		DB2 (DBPRINTF ("SliderBarGClass:   Notify target TOP\n"));

		NotifyAttrs ((Object *)g, msg->opu_GInfo,
			(msg->MethodID == OM_UPDATE) ? msg->opu_Flags : 0,
			PGA_Top,		sb->Top,
			GA_ID,			g->GadgetID,
			TAG_DONE);
	}

	return result;
}


static ULONG SB_OMGet (Class *cl, struct Gadget *g, struct opGet *msg)
{
	struct SBData *sb = INST_DATA (cl, g);

	ASSERT_VALID_PTR(sb)
	ASSERT_VALID_PTR(msg->opg_Storage)

	DB2 (DBPRINTF ("ScrollBarGClass: OM_GET\n");)


	switch (msg->opg_AttrID)
	{
		case PGA_Top:
			*msg->opg_Storage = (ULONG) sb->Top;
			return TRUE;

		case PGA_Total:
			*msg->opg_Storage = (ULONG) sb->Total;
			return TRUE;

		case PGA_Visible:
			*msg->opg_Storage = (ULONG) sb->Visible;
			return TRUE;

		default:
			return FALSE;
	}
}


static ULONG HOOKCALL SliderBarDispatcher (
	REG(a0, Class *cl),
	REG(a2, struct Gadget *g),
	REG(a1, Msg msg))

/* SliderBar class dispatcher - Handles all boopsi messages. */
{
	ASSERT_VALID_PTR(cl)
	ASSERT_VALID_PTR(g)
	ASSERT_VALID_PTR(msg)


	switch (msg->MethodID)
	{
		case GM_LAYOUT:
			/* This method is only supported on V39 and above */
			SB_GMLayout (cl, g, (struct gpLayout *)msg);
			return TRUE;

		case OM_SET:
		case OM_UPDATE:
			return SB_OMSet (cl, g, (struct opUpdate *)msg);

		case OM_GET:
			return SB_OMGet (cl, g, (struct opGet *)msg);

		case OM_NEW:
			return SB_OMNew (cl, g, (struct opSet *)msg);

		case OM_DISPOSE:
			SB_OMDispose (cl, g, msg);
			return TRUE;

		default:
			/* Unsupported method: let our superclass's dispatcher take
			 * a look at it. This includes all gadget methods sent
			 * by Intuition: GM_HITTEST, GM_RENDER, GM_HANDLEINPUT,
			 * GM_GOACTIVE and GM_GOINACTIVE. All these methods are
			 * automatically forwarded to our children by our superclass.
			 */
			DB2 (DBPRINTF ("SliderBarGClass: passing unknown method 0x%lx to superclass\n",
				msg->MethodID);)
			return DoSuperMethodA (cl, (Object *)g, msg);
	}
}



Class *MakeSliderBarGClass (void)
{
	Class *class;
	struct SBClassData *classdata;
	Class *smartgrp;

	if (!(smartgrp = MakeSmartGroupGClass()))
		return NULL;

	if (class = MakeClass (
#if (CLASS_FLAVOUR & FLAVOUR_PUBCLASS)
		SLIDERBARGCLASS,
#else
		NULL,
#endif
		NULL, smartgrp, sizeof (struct SBData), 0))
	{
		class->cl_Dispatcher.h_Entry = (ULONG (*)()) SliderBarDispatcher;

		/* Allocate storage for global class data */
		if (classdata = AllocMem (sizeof (struct SBClassData), MEMF_PUBLIC | MEMF_CLEAR))
		{
			class->cl_UserData = (ULONG) classdata;

			classdata->SmartGroupGClass = smartgrp;

			classdata->FrScrollButtonClass = MakeFrScrollButtonClass();
			ASSERT_VALID_PTR (classdata->FrScrollButtonClass);

			classdata->FrPropGClass = MakeFrPropGClass();
			ASSERT_VALID_PTR (classdata->FrPropGClass);

			if (classdata->VectorGlyphBase = (struct ClassLibrary *)
				OpenLibrary ("images/vectorglyph.image", 0))
			{
				classdata->VectorGlyphIClass = classdata->VectorGlyphBase->cl_Class;
				ASSERT_VALID_PTR (classdata->VectorGlyphIClass);

				if (classdata->Frame = NewObject (NULL, FRAMEICLASS,
					IA_EdgesOnly,	TRUE,
					IA_FrameType,	FRAME_BUTTON,
					TAG_DONE))
				{
#if (CLASS_FLAVOUR & FLAVOUR_PUBCLASS)
					AddClass (class);
#endif
					return class;
				}
			}
			else
			{
				static struct EasyStruct OpenClassES =
				{
					sizeof (struct EasyStruct),
					0,
#if (CLASS_FLAVOUR & FLAVOUR_CLASSLIB)
					LibName,
#else
					"sliderbar.gadget",
#endif
					"Couldn't open images/vectorglyph.image",
					"Ok"
				};

				/* Report an error */
				EasyRequestArgs (NULL, &OpenClassES, NULL, NULL);
			}
		}

		FreeSliderBarGClass (class);
	}

	return NULL;
}


BOOL FreeSliderBarGClass (Class *class)
{
	struct SBClassData *classdata;

	if (class)
	{
		ASSERT_VALID_PTR(class)

		classdata = (struct SBClassData *)class->cl_UserData;

		/* Try to remove the class */
		if (FreeClass (class))
		{
			if (classdata)
			{
				ASSERT_VALID_PTR(classdata)
				ASSERT_VALID_PTR_OR_NULL(classdata->Frame);
				ASSERT_VALID_PTR_OR_NULL(classdata->VectorGlyphBase)
				ASSERT_VALID_PTR_OR_NULL(classdata->FrScrollButtonClass)
				ASSERT_VALID_PTR_OR_NULL(classdata->FrPropGClass)
				ASSERT_VALID_PTR_OR_NULL(classdata->SmartGroupGClass)

				/* Cleanup global class data.
				 *
				 * NOTE: NULL is safe in DisposeObject() and CloseLibrary() since V36
				 */
				DisposeObject(classdata->Frame);
				CloseLibrary ((struct Library *)classdata->VectorGlyphBase);
				FreeFrScrollButtonClass (classdata->FrScrollButtonClass);
				FreeFrPropGClass (classdata->FrPropGClass);
				FreeSmartGroupGClass (classdata->SmartGroupGClass);
				FreeMem (classdata, sizeof (struct SBClassData));
			}
			return TRUE;
		}
		return FALSE;
	}
	return TRUE;
}



#if (CLASS_FLAVOUR & FLAVOUR_CLASSLIB)
/*
 * Class library support functions
 */

struct ClassLibrary * HOOKCALL _UserLibInit (REG(a6, struct ClassLibrary *mybase))
{
	SysBase = *((struct ExecBase **)4);	/* Initialize SysBase */

	IntuitionBase	= (struct IntuitionBase *) OpenLibrary ("intuition.library", 39);
	UtilityBase		= (UTILITYBASETYPE *) OpenLibrary ("utility.library", 39);

	if (!(IntuitionBase && UtilityBase))
	{
		_UserLibCleanup (mybase);
		return NULL;
	}

	if (mybase->cl_Class = MakeSliderBarGClass ())
	{
		return mybase;
	}
	else
	{
		_UserLibCleanup (mybase);
		return NULL;
	}
}

struct ClassLibrary * HOOKCALL _UserLibCleanup (REG(a6, struct ClassLibrary *mybase))
{
	if (mybase->cl_Class)
		if (!FreeSliderBarGClass (mybase->cl_Class))
			return NULL;

	CloseLibrary ((struct Library *)UtilityBase);
	CloseLibrary ((struct Library *)IntuitionBase);

	return mybase;
}

Class * HOOKCALL _GetEngine (REG(a6, struct ClassLibrary *mybase))
{
	return (mybase->cl_Class);
}

#endif /* (CLASS_FLAVOUR & FLAVOUR_CLASSLIB) */
