/*
**	$Id: SmartGroupGClass.c,v 1.1 1999/02/07 14:24:44 bernie Exp $
**
**	Copyright (C) 1999 Bernardo Innocenti
**	All rights reserved.
**
**	Use 4 chars wide TABs to read this file
**
**	"smartgroupgclass", a subclass of the gadgetclass which replaces
**	and enhances the builtin groupgclass.
**	See the header file "SmartGroupGClass.h" for class documentation.
*/

/* Definitions for system headers */
#define USE_BUILTIN_MATH
#define INTUI_V36_NAMES_ONLY
#define INTUITION_IOBSOLETE_H
#define __USE_SYSBASE
#define  CLIB_ALIB_PROTOS_H		/* Avoid dupe defines of boopsi funcs */

#include <exec/types.h>
#include <exec/memory.h>
#include <utility/tagitem.h>
#include <intuition/intuition.h>
#include <intuition/classes.h>
#include <intuition/classusr.h>
#include <intuition/gadgetclass.h>

#include <proto/exec.h>
#include <proto/intuition.h>
#include <proto/utility.h>

#ifdef __STORM__
	#pragma header
#endif

#include <gadgets/SmartGroupGClass.h>

#include <CompilerSpecific.h>
#include <DebugMacros.h>
#include <ListMacros.h>
#include <BoopsiStubs.h>
#include <BoopsiLib.h>


/* Local function prototypes */

static ULONG HOOKCALL SmartGroupDispatcher (REG(a0, Class *cl), REG(a2, Object *o), REG(a1, Msg msg));

static ULONG	ASMCALL SG_GMHitTest	(REG(a0, Class *cl), REG(a2, struct ExtGadget *g), REG(a1, struct gpHitTest *msg));
static void		ASMCALL SG_GMRender		(REG(a0, Class *cl), REG(a2, struct ExtGadget *g), REG(a1, struct gpRender *msg));
static ULONG	ASMCALL SG_GMGoActive	(REG(a0, Class *cl), REG(a2, struct ExtGadget *g), REG(a1, struct gpInput *msg));
static ULONG	ASMCALL SG_GMHandleInput(REG(a0, Class *cl), REG(a2, struct ExtGadget *g), REG(a1, struct gpInput *msg));
static void		ASMCALL SG_GMGoInactive (REG(a0, Class *cl), REG(a2, struct ExtGadget *g), REG(a1, struct gpGoInactive *msg));
static ULONG	ASMCALL SG_GMHelpTest	(REG(a0, Class *cl), REG(a2, struct ExtGadget *g), REG(a1, struct gpHitTest *msg));
static void		ASMCALL SG_GMLayout		(REG(a0, Class *cl), REG(a2, struct ExtGadget *g), REG(a1, struct gpLayout *msg));
static ULONG	ASMCALL SG_GMDomain		(REG(a0, Class *cl), REG(a2, struct ExtGadget *g), REG(a1, struct gpDomain *msg));
static ULONG	ASMCALL SG_OMNew		(REG(a0, Class *cl), REG(a2, struct ExtGadget *g), REG(a1, struct opSet *msg));
static void		ASMCALL SG_OMDispose	(REG(a0, Class *cl), REG(a2, struct ExtGadget *g), REG(a1, Msg msg));
//static ULONG	ASMCALL SG_OMSet		(REG(a0, Class *cl), REG(a2, struct ExtGadget *g), REG(a1, struct opUpdate *msg));
static void		ASMCALL SG_OMAddMember	(REG(a0, Class *cl), REG(a2, struct ExtGadget *g), REG(a1, struct opMember *msg));
static void		ASMCALL SG_OMRemMember	(REG(a0, Class *cl), REG(a2, struct ExtGadget *g), REG(a1, struct opMember *msg));

/* Per-object instance data */

struct GroupData
{
	struct MinList		ChildList;

	struct ExtGadget *	ActiveChild;
	struct IBox			GBox;

	/* Group orientation: can be LORIENT_VERT or LORIENT_HORIZ */
	ULONG				Orientation;
};



#if (CLASS_FLAVOUR & FLAVOUR_CLASSLIB)

	/* Class library support functions */
	struct ClassLibrary	*	HOOKCALL _UserLibInit		(REG(a6, struct ClassLibrary *mybase));
	struct ClassLibrary	*	HOOKCALL _UserLibCleanup	(REG(a6, struct ClassLibrary *mybase));
	Class *					HOOKCALL _GetEngine			(REG(a6, struct ClassLibrary *mybase));

	/* Library data */
	const UBYTE LibName[] = "sliderbar.gadget";
	const UBYTE LibVer[] = { '$', 'V', 'E', 'R', ':', ' ' };
	const UBYTE LibId[] = "smartgroup.gadget 1.0 (6.1.99)  1999 Bernardo Innocenti\n";

	/* Workaround a bug in StormC header file <proto/utility.h> */
	#ifdef __STORM__
		#define UTILITYBASETYPE struct Library
	#else
		#define UTILITYBASETYPE struct UtilityBase
	#endif

	/* Library bases */
	static struct ExecBase		*SysBase		= NULL;
	static struct IntuitionBase	*IntuitionBase	= NULL;
	static UTILITYBASETYPE		*UtilityBase	= NULL;

#endif



static ULONG ASMCALL SG_GMHitTest (REG(a0, Class *cl), REG(a2, struct ExtGadget *g), REG(a1, struct gpHitTest *msg))
{
	struct GroupData *gd = INST_DATA (cl, g);
	struct ExtGadget *child;
	APTR object_state;
	WORD x, y;

	DB2 (DBPRINTF ("SmartGroupGClass: GM_HITTEST\n");)
	ASSERT_VALID_PTR(gd)

	object_state = (APTR)gd->ChildList.mlh_Head;
	ASSERT_VALID_PTR(object_state)

	x = msg->gpht_Mouse.X + gd->GBox.Left;
	y = msg->gpht_Mouse.Y + gd->GBox.Top;

	/* Search over all our children */
	while (child = (struct ExtGadget *)NextObject(&object_state))
	{
		ASSERT_VALID_PTR(child)

		/* Check whether the mouse coordinates are inside
		 * the gadget rectangle of our child
		 */
		if ((x >= child->LeftEdge) &&
			(x < child->LeftEdge + child->Width) &&
			(y >= child->TopEdge) &&
			(y < child->TopEdge + child->Height))
		{
			struct gpHitTest childmsg;

			childmsg.MethodID		= GM_HITTEST;
			childmsg.gpht_GInfo		= msg->gpht_GInfo;
			childmsg.gpht_Mouse.X	= x - child->LeftEdge;
			childmsg.gpht_Mouse.Y	= y - child->TopEdge;

			if (DoMethodA ((Object *)child, (Msg)&childmsg) == GMR_GADGETHIT)
			{
				gd->ActiveChild = child;
				return GMR_GADGETHIT;
			}
		}
	}

	return 0;
}



static void ASMCALL SG_GMRender(REG(a0, Class *cl), REG(a2, struct ExtGadget *g), REG(a1, struct gpRender *msg))
{
	struct GroupData *gd = INST_DATA (cl, g);
	Object *o;
	APTR object_state;

	DB2 (DBPRINTF ("SmartGroupGClass: GM_RENDER\n");)

	ASSERT_VALID_PTR(gd)
	object_state = (APTR)gd->ChildList.mlh_Head;
	ASSERT_VALID_PTR(object_state)

	/* The groupgclass disposes all its children automatically */
	while (o = NextObject(&object_state))
	{
		ASSERT_VALID_PTR(o)
		DoMethodA (o, (Msg)msg);
	}
}



static ULONG ASMCALL SG_GMGoActive (REG(a0, Class *cl), REG(a2, struct ExtGadget *g), REG(a1, struct gpInput *msg))
{
	struct GroupData *gd = INST_DATA (cl, g);

	DB2 (DBPRINTF ("SmartGroupGClass: GM_GOACTIVE\n");)
	ASSERT_VALID_PTR(gd)
	ASSERT_VALID_PTR(gd->ActiveChild)

	if (gd->ActiveChild)
	{
		struct gpInput childmsg;

		childmsg.MethodID		= GM_GOACTIVE;
		childmsg.gpi_GInfo		= msg->gpi_GInfo;
		childmsg.gpi_IEvent		= msg->gpi_IEvent;
		childmsg.gpi_Termination= msg->gpi_Termination;
		childmsg.gpi_Mouse.X	= msg->gpi_Mouse.X + gd->GBox.Left
			- gd->ActiveChild->LeftEdge;
		childmsg.gpi_Mouse.Y	= msg->gpi_Mouse.Y + gd->GBox.Top
			- gd->ActiveChild->TopEdge;
		childmsg.gpi_TabletData	= msg->gpi_TabletData;

		return (DoMethodA ((Object *)gd->ActiveChild, (Msg)&childmsg));
	}

	gd->ActiveChild = NULL;
	return GMR_NOREUSE;
}



static ULONG ASMCALL SG_GMHandleInput (REG(a0, Class *cl), REG(a2, struct ExtGadget *g), REG(a1, struct gpInput *msg))
{
	struct GroupData *gd = INST_DATA (cl, g);

	DB2 (DBPRINTF ("SmartGroupGClass: GM_HANDLEINPUT\n");)
	ASSERT_VALID_PTR(gd)
	ASSERT_VALID_PTR(gd->ActiveChild)

	if (gd->ActiveChild)
	{
		struct gpInput childmsg;

		childmsg.MethodID		= GM_HANDLEINPUT;
		childmsg.gpi_GInfo		= msg->gpi_GInfo;
		childmsg.gpi_IEvent		= msg->gpi_IEvent;
		childmsg.gpi_Termination= msg->gpi_Termination;
		childmsg.gpi_Mouse.X	= msg->gpi_Mouse.X + gd->GBox.Left
			- gd->ActiveChild->LeftEdge;
		childmsg.gpi_Mouse.Y	= msg->gpi_Mouse.Y + gd->GBox.Top
			- gd->ActiveChild->TopEdge;
		childmsg.gpi_TabletData	= msg->gpi_TabletData;

		DB2 (DBPRINTF ("SmartGroupGClass:   Translating coords: X=%ld, Y=%ld\n",
			childmsg.gpi_Mouse.X, childmsg.gpi_Mouse.Y);)

		return DoMethodA ((Object *)gd->ActiveChild, (Msg)&childmsg);
	}

	return GMR_NOREUSE;
}



static void ASMCALL SG_GMGoInactive (REG(a0, Class *cl), REG(a2, struct ExtGadget *g), REG(a1, struct gpGoInactive *msg))
{
	struct GroupData *gd = INST_DATA (cl, g);

	DB2 (DBPRINTF ("SmartGroupGClass: GM_GOINACTIVE\n");)
	ASSERT_VALID_PTR(gd)
	ASSERT_VALID_PTR(gd->ActiveChild)

	if (gd->ActiveChild)
	{
		DoMethodA ((Object *)gd->ActiveChild, (Msg)msg);
		gd->ActiveChild = NULL;
	}
}



static ULONG ASMCALL SG_GMHelpTest (REG(a0, Class *cl), REG(a2, struct ExtGadget *g), REG(a1, struct gpHitTest *msg))
{
	struct GroupData *gd = INST_DATA (cl, g);
	struct ExtGadget *child;
	APTR object_state;
	WORD x, y;

	DB1 (DBPRINTF ("SmartGroupGClass: GM_HELPTEST\n");)

	ASSERT_VALID_PTR(gd)
	object_state = (APTR)gd->ChildList.mlh_Head;
	ASSERT_VALID_PTR(object_state)


	if ((g->Flags & GFLG_EXTENDED) && (g->MoreFlags & GMORE_GADGETHELP))
	{
		x = msg->gpht_Mouse.X + gd->GBox.Left;
		y = msg->gpht_Mouse.Y + gd->GBox.Top;

		/* Search over all our children */
		while (child = (struct ExtGadget *)NextObject(&object_state))
		{
			ASSERT_VALID_PTR(child)

			/* Limit to gadgets that understand GM_HELPTEST */
			if ((child->Flags & GFLG_EXTENDED) && (child->MoreFlags & GMORE_GADGETHELP))
			{
				if (child->MoreFlags & GMORE_BOUNDS)
				{
					/* Check whether the mouse coordinates are inside
					 * the gadget rectangle of our child
					 */
					if ((x >= child->BoundsLeftEdge) &&
						(x < child->BoundsLeftEdge + child->BoundsWidth) &&
						(y >= child->BoundsTopEdge) &&
						(y < child->BoundsTopEdge + child->BoundsHeight))
					{
						struct gpHitTest childmsg;

						childmsg.MethodID		= GM_HELPTEST;
						childmsg.gpht_GInfo		= msg->gpht_GInfo;
						childmsg.gpht_Mouse.X	= x - child->BoundsLeftEdge;
						childmsg.gpht_Mouse.Y	= y - child->BoundsTopEdge;

						if (DoMethodA ((Object *)child, (Msg)&childmsg) == GMR_HELPHIT)
						{
							gd->ActiveChild = child;
							return GMR_HELPHIT;
						}
					}
				}
				else /* No valid bounds, use gadget select box */
				{
					/* Check whether the mouse coordinates are inside
					 * the gadget rectangle of our child
					 */
					if ((x >= child->LeftEdge) &&
						(x < child->LeftEdge + child->Width) &&
						(y >= child->TopEdge) &&
						(y < child->TopEdge + child->Height))
					{
						struct gpHitTest childmsg;

						childmsg.MethodID		= GM_HELPTEST;
						childmsg.gpht_GInfo		= msg->gpht_GInfo;
						childmsg.gpht_Mouse.X	= x - child->LeftEdge;
						childmsg.gpht_Mouse.Y	= y - child->TopEdge;

						if (DoMethodA ((Object *)child, (Msg)&childmsg) == GMR_HELPHIT)
						{
							gd->ActiveChild = child;
							return GMR_HELPHIT;
						}
					}
				}
			}
		}
	}

	return GMR_NOHELPHIT;
}



static void ASMCALL SG_GMLayout(REG(a0, Class *cl), REG(a2, struct ExtGadget *g), REG(a1, struct gpLayout *msg))
{
	struct GroupData *gd = (struct GroupData *) INST_DATA (cl, (Object *)g);
	struct ExtGadget *child;
	APTR object_state;

	DB2 (DBPRINTF ("SmartGroupGClass: GM_LAYOUT\n");)
	ASSERT_VALID_PTR(gd)

	/* Collect new group size */
	GetGadgetBounds(msg->gpl_GInfo, g, &gd->GBox);

	/* Forward GM_LAYOUT to our children that need it */

	object_state = (APTR)gd->ChildList.mlh_Head;
	ASSERT_VALID_PTR(object_state)

	while (child = (struct ExtGadget *)NextObject(&object_state))
	{
		ASSERT_VALID_PTR(child)

		if (child->Flags & (
			GFLG_RELBOTTOM |
			GFLG_RELRIGHT |
			GFLG_RELWIDTH |
			GFLG_RELHEIGHT |
			GFLG_RELSPECIAL))
		{
			DB2 (DBPRINTF ("SmartGroupGClass:   Forwarding GM_LAYOUT to child\n");)
			DoMethodA ((Object *)child, (Msg)msg);
		}
	}
}



static ULONG ASMCALL SG_GMDomain (REG(a0, Class *cl), REG(a2, struct ExtGadget *g), REG(a1, struct gpDomain *msg))
{
	DB1 (DBPRINTF ("SmartGroupGClass: GM_DOMAIN\n");)

	/* TODO */
	return 0;
}



static ULONG ASMCALL SG_OMNew (REG(a0, Class *cl), REG(a2, struct ExtGadget *g), REG(a1, struct opSet *msg))
{
	struct GroupData *gd;
	struct TagItem *ti, *tstate = msg->ops_AttrList;

	DB1 (DBPRINTF ("SmartGroupGClass: OM_NEW\n");)

	if (g = (struct ExtGadget *)DoSuperMethodA (cl, (Object *)g, (Msg)msg))
	{
		gd = (struct GroupData *) INST_DATA (cl, (Object *)g);
		ASSERT_VALID_PTR(gd)

		DB (if (g->Flags & GFLG_EXTENDED)
			DBPRINTF ("SmartGroupGClass:   Wow, we are an ExtGadget!\n");)

		/* Initialize object instance data */
		NEWLIST((struct List *)&gd->ChildList);
		gd->ActiveChild = NULL;
		gd->ActiveChild = NULL;

		gd->Orientation = LORIENT_NONE;

		while (ti = NextTagItem (&tstate))
			switch (ti->ti_Tag)
			{
				case SGA_Child:
				{
					struct opMember opam;

					if (!ti->ti_Data)
					{
						/* Commit suicide without disturbing the
						 * dispatchers of our subclasses
						 */
						SG_OMDispose (cl, g, NULL);
						return NULL;
					}

					ASSERT_VALID_PTR(ti->ti_Data)
					opam.opam_Object = (Object *)ti->ti_Data;
					SG_OMAddMember (cl, g, &opam);
					break;
				}

				case LAYOUTA_Orientation:
					gd->Orientation = ti->ti_Data;
					break;

				default:
					break;
			}

		/* Intuition sends GM_LAYOUT messages only to gadgets with
		 * relative postion. Since we need to recalculate the
		 * positions of our children when the window resizes, we
		 * set the GFLG_RELSPECIAL property if to be notified in all cases.
		 */
		g->Flags |= GFLG_RELSPECIAL;
	}

	return (ULONG)g;
}



static void ASMCALL SG_OMDispose (REG(a0, Class *cl), REG(a2, struct ExtGadget *g), REG(a1, Msg msg))
{
	struct GroupData *gd = (struct GroupData *) INST_DATA (cl, (Object *)g);
	Object *o;
	APTR object_state;

	DB1 (DBPRINTF ("SmartGroupGClass: OM_DISPOSE\n");)

	ASSERT_VALID_PTR(gd)
	object_state = (APTR)gd->ChildList.mlh_Head;
	ASSERT_VALID_PTR(object_state)

	/* The groupgclass disposes all its children automatically */
	while (o = NextObject(&object_state))
	{
		ASSERT_VALID_PTR(o)
		DoMethod (o, OM_REMOVE);
		DisposeObject(o);
	}

	/* Our superclass will cleanup everything else now */
	DoSuperMethodA (cl, (Object *)g, msg);

	/* From now on, our instance data is no longer available */
}


#if 0
static ULONG ASMCALL SG_OMSet (REG(a0, Class *cl), REG(a2, struct ExtGadget *g), REG(a1, struct opUpdate *msg))
{
	struct GroupData *gd = INST_DATA (cl, g);
	struct TagItem *ti, *tstate = msg->opu_AttrList;

	DB1 (DBPRINTF ("SmartGroupGClass: OM_SET\n");)

	ASSERT_VALID_PTR(gd)
	ASSERT_VALID_PTR_OR_NULL(tstate)

	DB2 (DBPRINTF ((msg->MethodID == OM_SET) ?
		"SliderBarGClass: OM_SET\n" :
		"SliderBarGClass: OM_UPDATE\n");)


	while (ti = NextTagItem (&tstate))
		switch (ti->ti_Tag)
		{
			default:
				break;
		}

		result = TRUE;


	return result;
}
#endif



static void ASMCALL SG_OMAddMember (REG(a0, Class *cl), REG(a2, struct ExtGadget *g), REG(a1, struct opMember *msg))
{
	struct GroupData *gd = INST_DATA (cl, g);
	ASSERT_VALID_PTR(gd)
	DB1 (DBPRINTF ("SmartGroupGClass: OM_ADDMEMBER\n");)

	if(msg->opam_Object)
	{
		ASSERT_VALID_PTR(msg->opam_Object)

		/* TODO: check for GMORE_SCROLLRASTER ? */

		DoMethod (msg->opam_Object, OM_ADDTAIL, &gd->ChildList);
	}
}



static void ASMCALL SG_OMRemMember (REG(a0, Class *cl), REG(a2, struct ExtGadget *g), REG(a1, struct opMember *msg))
{
	struct GroupData *gd = INST_DATA (cl, g);

	DB1 (DBPRINTF ("SmartGroupGClass: OM_REMMEBRER\n");)
	ASSERT_VALID_PTR(gd)

	DoMethod (msg->opam_Object, OM_REMOVE);

	if (msg->opam_Object == (Object *)gd->ActiveChild)
		gd->ActiveChild = NULL;

	/* TODO: resize & redraw group */
}



static ULONG HOOKCALL SmartGroupDispatcher (
	REG(a0, Class *cl),
	REG(a2, Object *o),
	REG(a1, Msg msg))

/* SmartGroup class dispatcher - Handles all boopsi messages. */
{
	static const HookPtr GMHooks[] =
	{
		/* GM_HITTEST		*/ (HookPtr)SG_GMHitTest,
		/* GM_RENDER		*/ (HookPtr)SG_GMRender,
		/* GM_GOACTIVE		*/ (HookPtr)SG_GMGoActive,
		/* GM_HANDLEINPUT	*/ (HookPtr)SG_GMHandleInput,
		/* GM_GOINACTIVE	*/ (HookPtr)SG_GMGoInactive,
		/* GM_HELPTEST		*/ (HookPtr)SG_GMHelpTest,
		/* GM_LAYOUT		*/ (HookPtr)SG_GMLayout,
		/* GM_DOMAIN		*/ (HookPtr)SG_GMDomain
	};

	static const HookPtr OMHooks[] =
	{
		/* OM_NEW		*/ (HookPtr)SG_OMNew,
		/* OM_DISPOSE	*/ (HookPtr)SG_OMDispose,
		/* OM_SET		*/ (HookPtr)NULL,
		/* OM_GET		*/ (HookPtr)NULL,
		/* OM_ADDTAIL	*/ (HookPtr)NULL,
		/* OM_REMOVE	*/ (HookPtr)NULL,
		/* OM_NOTIFY	*/ (HookPtr)NULL,
		/* OM_UPDATE	*/ (HookPtr)NULL,
		/* OM_ADDMEMBER	*/ (HookPtr)SG_OMAddMember,
		/* OM_REMMEMBER	*/ (HookPtr)SG_OMRemMember
	};

	ULONG id;

	ASSERT_VALID_PTR(cl)
	ASSERT_VALID_PTR(o)
	ASSERT_VALID_PTR(msg)

	/* Optimized boopsi class dispatcher.
	 *
	 * This class overrides almost all gadgetclass and rootclass messages.
	 * The usual switch() statement on msg->MessageID would be quite inefficient.
	 * Most compilers aren't smart enough to implement it with two jump tables
	 * as we did here. This code relies on the values assigned to the method IDs
	 * in the headers to remain constant. It will break otherwise.
	 */

	id = msg->MethodID;

	if (id <= GM_DOMAIN)
		return GMHooks[id](cl, o, msg);

	id -= OM_NEW;

	if (id <= (OM_REMMEMBER - OM_NEW))
		if (OMHooks[id])
			return OMHooks[id](cl, o, msg);

	return DoSuperMethodA (cl, o, msg);
}



Class *MakeSmartGroupGClass (void)
{
	Class *class;

	if (class = MakeClass (
#if (CLASS_FLAVOUR & FLAVOUR_PUBCLASS)
		SMARTGROUPGCLASS,
#else
		NULL,
#endif
		GADGETCLASS, NULL, sizeof (struct GroupData), 0))
	{
		class->cl_Dispatcher.h_Entry = (ULONG (*)()) SmartGroupDispatcher;
	}

	return class;
}


BOOL FreeSmartGroupGClass (Class *class)
{
	if (class)
	{
		ASSERT_VALID_PTR(class)

		/* Try to remove the class */
		if (FreeClass (class))
			return TRUE;

		return FALSE;
	}
	return TRUE;
}



#if (CLASS_FLAVOUR & FLAVOUR_CLASSLIB)
/*
 * Class library support functions
 */

struct ClassLibrary * HOOKCALL _UserLibInit (REG(a6, struct ClassLibrary *mybase))
{
	SysBase = *((struct ExecBase **)4);	/* Initialize SysBase */

	IntuitionBase	= (struct IntuitionBase *) OpenLibrary ("intuition.library", 39);
	UtilityBase		= (UTILITYBASETYPE *) OpenLibrary ("utility.library", 39);

	if (!(IntuitionBase && UtilityBase))
	{
		_UserLibCleanup (mybase);
		return NULL;
	}

	if (mybase->cl_Class = MakeSmartGroupGClass ())
	{
		return mybase;
	}
	else
	{
		_UserLibCleanup (mybase);
		return NULL;
	}
}

struct ClassLibrary * HOOKCALL _UserLibCleanup (REG(a6, struct ClassLibrary *mybase))
{
	if (mybase->cl_Class)
		if (!FreeSliderBarGClass (mybase->cl_Class))
			return NULL;

	CloseLibrary ((struct Library *)UtilityBase);
	CloseLibrary ((struct Library *)IntuitionBase);

	return mybase;
}

Class * HOOKCALL _GetEngine (REG(a6, struct ClassLibrary *mybase))
{
	return (mybase->cl_Class);
}

#endif /* (CLASS_FLAVOUR & FLAVOUR_CLASSLIB) */
