/*
**	$Id: SmartGroupGClass.h,v 1.1 1999/02/07 14:24:44 bernie Exp $
**
**	Copyright (C) 1999 Bernardo Innocenti
**	All rights reserved.
**
**	"smartgroupgclass", a class similar to the groupgclass
**	that allows attaching any number of child gadgets that
**	will behave like a single gadget. Input processing and
**	rendering are deferred to the children as appropriate.
**
**	The original ROM class is not very usable because it has
**	a lot of bugs and limitations:
**
**	- Doesn't forward GM_LAYOUT to its children
**	- Does not work when GA_RelRight or GA_RelBottom are set
**	- Ignores GA_(Rel)Width and GA_(Rel)Height on creation
**	  and recalculates its size when new children are attached.
**	- Can not layout its children automatically
**
**	The smartgroupgclass provides the custom creation-time
**	attribute SGA_Child, which allows to attach child gadgets
**	to it in a single call to NewObject() instead of calling
**	OM_ADDMEMBER several times later. This enables the user
**	to create a whole GUI with a single C statement like this:
**
**	if (glist = NewObject(NULL, SMARTGROUPGCLASS,
**		LAYOUTA_Orientation, LORIENT_VERT,
**		SGA_Child, NewObject(NULL SMARTGROUPGCLASS,
**			LAYOUTA_Orientation, LORIENT_HORIZ,
**			SGA_Child, NewObject(...),
**			SGA_Child, NewObject(...),
**			TAG_DONE),
**		SGA_Child, NewObject(...),
**		TAG_DONE))
**		{
**			/* ... use your glist here... */
**			DeleteObject (glist);
**		}
**		else
**			printf ("Failed to create the application!\n");
*/

#define SMARTGROUPGCLASS "smartgroupgclass"
#define SMARTGROUPVERS	1


/* Class initializer and terminator. Only avalable when
 * the class is compiled in the static linker object flavour.
 */
Class	*MakeSmartGroupGClass (void);
BOOL	 FreeSmartGroupGClass (Class *class);


/*
 * New Methods:
 *
 *	This class does not define any new methods
 */


/*
 * New Attributes:
 *
 */
#define SGA_Dummy			(TAG_USER | ('S'<<16) | ('G'<<8))

#define SGA_Child			(SGA_Dummy + 1)
	/* [I] (struct Gadget *) - Attach the specified child to
	 * the group at creation time. If the argument is NULL,
	 * the group will be disposed and OM_NEW will also
	 * return NULL. This makes it easy to create a whole user
	 * interface by nesting several calls to NewObject() and
	 * testing the result of the higher level group only.
	 */

/*
 * Changed attributes:
 *
 * LAYOUTA_Orientation
 *	[I] (ULONG) - Pass one of the LORIENT_#? modes.
 *	Defaults to LORIENT_HORIZ.
 *
 */

/* FIX: Define the attributes that not present
 * before V42 of <intuition/gadgetclass.h>.
 */
#ifndef LAYOUTA_Dummy
	#define	LAYOUTA_Dummy 		(TAG_USER+0x38000)
	#define LAYOUTA_Orientation	(LAYOUTA_Dummy+0x0003)
	#define LORIENT_NONE	0
	#define LORIENT_HORIZ	1
	#define LORIENT_VERT	2
#endif


/*
 * Changed methods:
 *
 *	The class redefines all the GM_#? methods and defers processing
 *	to its children in a way that makes the group behave as it was
 *	a single gadget. The children may still choose to send
 *	notification directly to the application or to gadgets attached
 *	to other groups.
 */
